# This handles parsing of rental homes data
from typing import Dict, Any, List
from numbers import Number
from .details_page import Preload_Detail_Page


class Rental_Home(Preload_Detail_Page):
    """This class extracts the properties/details of a rental property from Zillow's detail URL page. 

    While most of the key details can be easily access via the class attributes or functions, there are some more advanced details that not parsed/organized by this class. For the full html page, see the soup attribute. Additionally check the variant_cache and the full_cache for the complete dictionaries of the api cache. 

    Attributes:
        url (str): The detail URL that this class has parsed
        soup (BeautifulSoup): The html content of this page parsed to a soup object
        preload (Dict[Any, Any]): The complete api cache preload

        zpid (int): The Zillow property ID

        variant_cache (Dict[Any, Any]): The variant api cache
        full_cache (Dict[Any, Any]): The full api cache

        property (Dict[Any, Any]): The property data dictionary

        status (str) : The home status, for example: FOR_SALE or FOR_RENT
        home_type (str): The home type
        year_built (int): Year home was built

        price (int): The monthly rent
        zestimate (int): Zillow's estimated home value
        rental_zestimate (int): Zillow's estimated monthly rental value
        tax_history (List[Dict[str, Number]]): The tax history of the property
        price_history (List[Dict[str, Any]]): The price history of the property
        currency (str): The currency the price, zestimate, etc. are in

        status (str): The status of the listing. For example - for sale.
        days_on_zillow (int): The number of days on Zillow
        views (int): The number of views on Zillow
        saves (int): The number of saves/favorites on Zillow

        tags (List[str]): The tags on the property. For example - "Rolling Hills".
        description (str): The property description

        address (Dict[str, Any]): Dictionary of data related to the address of the property
        street_address (str): The street address
        city (str): City
        state (str): State
        zip (str): zip code
        latitude (float): The latitude
        longitude (float): The longitude

        bedrooms (Number): The number of bedrooms
        bathrooms (Number): The number of bathrooms

        interior_sqft (Number): The SQFT area of the interior of the property

        appliances (List[str]): The list of appliances that are included with the property
        cooling (List[str]): The list of cooling related features 
        heating (List[str]): The list of heating related features
        community_features (List[str]): The list of community features
        fireplaces (Number): The number of fireplaces. Note: this can often be incorrect or misleading
        garage (bool): True if has garage, False if doesn't
        interior_features (List[str]): List of interior features

        attic (str | None): None if has no attic, otherwise will have a string specification of attic type
        basement (str): Specification of basement type

        hoa_fee (Number | None): If has a hoa fee, will be the monthly fee as a number. If not, then will be None

        levels (str): The number of levels given as a string

        parking (List[str]): The parking features, for example ["OFF_STREET"]
        lot_features (List[str] | None): List of lot features, if there are none, will be None
        lot_size (str): Lot size as as a string with units
        lot_size_dimensions (str): Dimensions of the lot as a string
        lot_sqft (Number): The SQFT area of the lot

        sewer (List[str]): The list of sewer systems. Typically this will just be a list of one element.
        water_source (List[str]): The list of water sources. Typically this will just be a list of one element.

        attribution (Dict[str, Any]): The attribution details of the listing agent/agency

        schools (List[Dict[str, Any]]): The nearby schools
        similar (List[Dict[str, Any]]): Similar properties
        nearby (List[Dict[str, Any]]): Nearby properties

        fees_and_dues (List[Dict[str, Any]]): Fees/dues associated with property
    """

    def __init__(self, url: str) -> None:
        """This initializes the Rental_Home object, which call a GET request on the Zillow detail URL. 

        Args:
            url (str): The Zillow Rental Property details URL.
        """

        super().__init__(url)

        self.fees_and_dues: List[Dict[str, Any]
                                 ] = self.property['resoFacts']['feesAndDues']



HEADER = {
    "authority": "www.landwatch.com",
    "accept": "*/*",
    "accept-language": "en-US,en;q=0.7",
    "referer": "https://www.landwatch.com/tennessee-land-for-sale/chattanooga",
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "cors",
    "sec-fetch-site": "same-origin",
    "sec-gpc": "1",
    "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36"
}

ROOT_URL = "https://www.landwatch.com/api/property/search/1113"



# URL for graphQL listing serach requests
SEARCH_URL = "https://www.realtor.com/api/v1/hulk_main_srp"

HEADER = {
    "authority": "www.realtor.com",
    "accept": "*/*",
    "accept-language": "en-US,en;q=0.9",
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "cors",
    "sec-fetch-site": "same-origin",
    "sec-gpc": "1",
    "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36"
}

GRAPHQL_HEADER = {
    "authority": "www.realtor.com",
    "accept": "application/json",
    "accept-language": "en-US,en;q=0.9",
    "content-type": "application/json",
    "origin": "https://www.realtor.com",
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "cors",
    "sec-fetch-site": "same-origin",
    "sec-gpc": "1",
    "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36"
}

# This is the GRAPHQL query that is used by Realtor.com
# Note that it requires several variables
# query --> search query, with variables such as status, search_location, etc.
# client_data --> the client data
# limit --> the result limit (if 42, will limit to 42 results)
# offset --> the result offset, so if 42, will start from result 42 (43rd result counting from 0)
# sort --> unclear what this is, realtor.com doesn't seem to use this variable
# sort_type --> the way in which to sort, by default "relevant"
# bucket --> unclear what this is, realtor.com doesn't seem to use this variable
# operationName --> the name of the operation
GRAPHQL_LISTING_SEARCH_QUERY = """
query ConsumerSearchMainQuery(
  $query: HomeSearchCriteria!
  $limit: Int
  $offset: Int
  $sort: [SearchAPISort]
  $sort_type: SearchSortType
  $client_data: JSON
  $bucket: SearchAPIBucket
) {
  home_search: home_search(
    query: $query
    sort: $sort
    limit: $limit
    offset: $offset
    sort_type: $sort_type
    client_data: $client_data
    bucket: $bucket
  ) {
    count
    total
    results {
      property_id
      list_price
      primary
      primary_photo(https: true) {
        href
      }
      source {
        id
        agents {
          office_name
        }
        type
        spec_id
        plan_id
      }
      community {
        property_id
        description {
          name
        }
        advertisers {
          office {
            hours
            phones {
              type
              number
            }
          }
          builder {
            fulfillment_id
          }
        }
      }
      products {
        brand_name
        products
      }
      listing_id
      matterport
      virtual_tours {
        href
        type
      }
      status
      permalink
      price_reduced_amount
      other_listings {
        rdc {
          listing_id
          status
          listing_key
          primary
        }
      }
      description {
        beds
        baths
        baths_full
        baths_half
        baths_1qtr
        baths_3qtr
        garage
        stories
        type
        sub_type
        lot_sqft
        sqft
        year_built
        sold_price
        sold_date
        name
      }
      location {
        street_view_url
        address {
          line
          postal_code
          state
          state_code
          city
          coordinate {
            lat
            lon
          }
        }
        county {
          name
          fips_code
        }
      }
      tax_record {
        public_record_id
      }
      lead_attributes {
        show_contact_an_agent
        opcity_lead_attributes {
          cashback_enabled
          flip_the_market_enabled
        }
        lead_type
        ready_connect_mortgage {
          show_contact_a_lender
          show_veterans_united
        }
      }
      open_houses {
        start_date
        end_date
        description
        methods
        time_zone
        dst
      }
      flags {
        is_coming_soon
        is_pending
        is_foreclosure
        is_contingent
        is_new_construction
        is_new_listing(days: 14)
        is_price_reduced(days: 30)
        is_plan
        is_subdivision
      }
      list_date
      last_update_date
      coming_soon_date
      photos(limit: 2, https: true) {
        href
      }
      tags
      branding {
        type
        photo
        name
      }
    }
  }
}
"""

# This is the request payload for the listing search GRAPHQL API
# This query setting two fields, 'query' (graphQL query) and 'variables' --> 'query' (which is the filter criteria for the GRAPHQL request)
LISTING_SEARCH_PAYLOAD = {
    "query": "",
    "variables": {
        "query": {},
        "client_data": {"device_data": {"device_type": "web"}, "user_data": {}},
        "limit": 42,
        "offset": 42,
        "sort_type": "relevant",
    },
    "operationName": "ConsumerSearchMainQuery",
    "callfrom": "SRP",
    "nrQueryType": "MAIN_SRP",
    "isClient": True,
}



import json
import requests
from bs4 import BeautifulSoup
from numbers import Number
from typing import Dict, Any, Tuple, List
from datetime import datetime
from .. import defaults


class Sale:
    """This class extracts the properties/details of for sale property from Realtor.com's detail URL page.

    While most of the key details can be easily accessed via the class attribute or functions, there are some bit of data that are not readily assigned to class attributes. This includes details of appliances, which can be found in the details attribute (Dict[str, Any]).

    Attribute:
        url (str): The detail URL that this class has parsed
        soup (BeautifulSoup): The html content of this page parsed to a soup object

        ndata (Dict): The __NEXT_DATA__ data dictionary pulled from the html soup
        initial_state (Dict): The initial state dictionary component of ndata
        property_details (Dict): propertyDetails component of the initial_state dictionary
        description (Dict): description component of the property_details dictionary. Contains key value pairs regarding to the description of the property
        location (Dict): location component of the property_details dictionary. Contains key value pairs regarding to the location of the property

        property_id (str): The realtor.com property id
        listing_date (str): The date the property was listed

        status (str): The status of the property
        price (int): The price of the property, in USD
        price_per_sqft (int): The price per sqft, rounded, in USD
        yearly_property_tax (Number): The yearly property tax - how much tax is paid on the property each year
        open_houses (Any): The open house information

        listing_description (str): The text description of the property
        details (Dict[str, Any]): The details of the property, including appliances, cooling, septic, etc.
        beds (int): Number of bedrooms
        baths (int): Number of bathrooms
        garage (int): The number of garages/number of car spaces that fit in the garage
        interior_sqft (int): The total sqft area of the interior
        lot_sqft (int): The total sqft area of the lot

        address (Dict[str, Any]): Data dictionary of key value pairs relating to the property's address
        city (str): The city 
        state_code (str): The state code
        state (str): The state
        county (str): The county
        latitude (float): The latitude coordinate of the property
        longitude (float): The longitude coordinate of the property
        zip (str): The zip code
        street_address (str): The street address
        fips (str): Federal Information Processing System code

        hoa_fee (int): The hoa fee (0 if none)

        property_history (List[Dict]): The property history (include sales, listing added, removed, etc.)
        tax_history (List[Dict]): The tax history of the property

        area_market_status (Dict): Realtor.com's description of the area's market status (for example it may state the housing market is 'hot')

        noise (str): The noise level status in the area

        schools (Dict): Nearby schools
        similar (Dict): Similar homes, according to Realtor.com

    """

    def __init__(self, url: str) -> None:
        """This initializes the Preload_Detail_Page object, which call a GET request on the Zillow detail URL. 

        Args:
            url (str): The Zillow Property Sale details URL.
        """

        self.url = url

        # get soup
        self.soup = self.make_soup(self.get_page(self.url))

        # get ndata
        self.ndata = self.get_next_data(self.soup)

        # get initialState and propertyDetails
        self.initial_state: Dict[str,
                                 Any] = self.ndata['props']['pageProps']['initialState']
        self.property_details: Dict[str,
                                    Any] = self.initial_state['propertyDetails']

        # additional important keys
        self.description: Dict[str, Any] = self.property_details['description']
        self.location: Dict[str, Any] = self.property_details['location']

        # quick access values

        self.property_id: str = self.property_details['property_id']
        self.listing_date: str = self.property_details['list_date']

        self.status: str = self.property_details['status']
        self.price: int = self.property_details['list_price']
        self.price_per_sqft: int = self.property_details['price_per_sqft']
        self.yearly_property_tax: Number = self.property_details['source']['raw']['tax_amount']
        self.year_built: int = self.description['year_built']
        self.open_houses: Any = self.property_details['open_houses']

        self.listing_description: str = self.description['text']
        self.details: Dict[str, Any] = self.property_details['details']
        self.beds: int = self.description['beds']
        self.baths: int = self.description['baths']
        self.garage: int = self.description['garage']
        self.interior_sqft: int = self.description['sqft']
        self.lot_sqft: int = self.description['lot_sqft']

        self.address: Dict[str, Any] = self.location['address']
        self.city: str = self.address['city']
        self.state_code: str = self.address['state_code']
        self.state: str = self.address['state']
        self.county: str = self.address['county']
        self.latitude: float = self.address['coordinate']['lat']
        self.longitude: float = self.address['coordinate']['lon']
        self.zip: str = self.address['postal_code']
        self.street_address = f"{self.address['line']}, {self.city}, {self.state_code} {self.zip}"
        self.fips: str = self.location['county']['fips_code']

        self.hoa_fee: int = self.property_details['hoa'].get('fee', 0)

        self.property_history: List[Dict[str, Any]
                                    ] = self.property_details['property_history']
        self.tax_history: List[Dict[str, Any]
                               ] = self.property_details['tax_history']

        self.area_market_status: Dict[str, Any] = self.location.get('postal_code', {}).get(
            'geo_statistics', {}).get('housing_market')

        self.noise: str = self.get_noise_metrics(
            self.latitude, self.longitude).get('local_text', 'Unknown')

        self.schools: Dict[str, Any] = self.property_details['schools']
        self.similar = self.get_similar_homes(self.property_id)

    @staticmethod
    def get_page(url: str, headers: Dict = defaults.HEADER) -> requests.Response:
        """Submits a GET request to the detail URL to get the page.

        Args:
            url (str): The detail URL. Note that invalid URL will lead to unexpected results and errors
            headers (Dict, optional): The request headers. Incorrect headers may lead to invalid results. It is recommended to leave as default. Defaults to defaults.HEADER.

        Returns:
            requests.Response: The page GET response
        """
        return requests.get(url, headers=headers)

    @staticmethod
    def make_soup(page: requests.Response) -> BeautifulSoup:
        """This is a very simple function that has barely a reason to exist. Just takes the response content and has it parsed by Beautiful Soup via its html parser.

        Args:
            page (requests.Response): The GET response from a valid GET requests on the detail URL page

        Returns:
            BeautifulSoup: Page HTML parsed as a BeautifulSoup Object
        """

        soup = BeautifulSoup(page.content, "html.parser")

        return soup

    @staticmethod
    def get_next_data(soup: BeautifulSoup) -> Dict[Any, Any]:
        """Extracts and parses the NEXT_DATA cache from the page soup.

        Args:
            soup (BeautifulSoup): The html page soup object

        Returns:
            Dict[Any, Any]: The api preload cache dictionary
        """

        ndata = soup.find("script", id="__NEXT_DATA__").text
        ndata = json.loads(ndata)

        return ndata

    @staticmethod
    def get_loan_estimates(
        home_price: int,
        down_payment: int,
        fips: str,
        state_code: str,
        yearly_property_tax: int | float,
        hoa_fee: int = 0,
    ) -> Dict[str, Any]:
        """Gets the loan estimates from Realtor.com

        Args:
            home_price (int): The home price
            down_payment (int): The downpayment amount. Note that Realtor.com typically uses 20%, smaller downpayments may not be handled by Realtor.com's API correctly.
            fips (str): The FIPS (Federal Information Processing System) code 
            state_code (str): The state code
            yearly_property_tax (int | float): The yearly property tax ammount
            hoa_fee (int, optional): The hoa fee. Defaults to 0.

        Returns:
            Dict[str, Any]: Returns the API response json
        """

        url = "https://www.realtor.com/api/v1/payments/calculate_property_loan"

        querystring = {
            "hoa_fees": hoa_fee,
            "fips": fips,
            "state": state_code,
            "home_price": home_price,
            "down_payment": down_payment,
            "veterans_benefits": "false",
            "property_tax": yearly_property_tax,
            "is_fees_included": "true",
            "app_name": "realtor_dot_com",
            "app_version": "0.0.1"
        }

        response = requests.request("GET", url, params=querystring)
        return response.json()['results']

    def get_estimated_monthly_payment(self, down_payment: int = None) -> Number:
        """Gets the estimated monthly payment.

        Args:
            down_payment (int, optional): The downpayment on the purchase. If set to None, will default to 20% of the home price. Defaults to None.

        Returns:
            Number: Returns the monthly payment estimate
        """

        if down_payment is None:
            down_payment = round(self.price * .2)

        results = self.get_loan_estimates(
            self.price,
            down_payment,
            self.fips,
            self.state_code,
            self.yearly_property_tax,
            self.hoa_fee
        )

        return results['mortgage_data']['monthly_payment']

    @staticmethod
    def get_noise_metrics(latitude: float, longitude: float) -> Dict[str, Any]:
        """Gets the noise metrics for the coordinates specified.

        Args:
            latitude (float): The latitude
            longitude (float): The longitude

        Returns:
            Dict[str, Any]: The noise metrics
        """

        noise_query = f"https://www.realtor.com/api/v1/maps/gstat/noise?lat={latitude}&lon={longitude}"

        response = requests.get(noise_query).json()

        return response['result']

    @staticmethod
    def get_flood_risk(property_id: str, headers: Dict = defaults.HEADER) -> Dict[str, Any]:
        """Gets the flood risk from Realtor.com's PUBLIC API

        Args:
            property_id (str): The Realtor.com property ID
            headers (Dict, optional): The request headers. Incorrect headers may lead to invalid results. It is recommended to leave as default. Defaults to defaults.HEADER.

        Returns:
            Dict[str, Any]: The flood risk data
        """

        url = "https://www.realtor.com/api/v1/hulk"

        querystring = {"client_id": "rdc-x", "schema": "vesta"}

        payload = {
            "query": "query GetLocalData($propertyId: ID!) {  home(property_id: $propertyId) {    local {      flood {        fsid        flood_factor_score        flood_factor_severity        flood_cumulative_30        flood_trend        flood_trend_paragraph        fema_zone        firststreet_url        flood_insurance_text        environmental_risk      trend_direction      insurance_requirement      insurance_rates {          provider_logo          provider_url          providers        }        insurance_quotes {          provider_name          provider_url          provider_logo          expires          price          home_coverage          contents_coverage          disclaimer        }      }    }  }}",
            "variables": {"propertyId": property_id}
        }

        response = requests.request(
            "POST", url, json=payload, headers=headers, params=querystring)
        return response.json()

    @staticmethod
    def get_fire_risk(property_id: str, headers: Dict = defaults.HEADER) -> Dict[str, Any]:
        """Gets the fire risk from Realtor.com's PUBLIC API

        Args:
            property_id (str): The Realtor.com property ID
            headers (Dict, optional): The request headers. Incorrect headers may lead to invalid results. It is recommended to leave as default. Defaults to defaults.HEADER.

        Returns:
            Dict[str, Any]: The fire risk data
        """

        url = "https://www.realtor.com/api/v1/hulk"

        querystring = {"client_id": "rdc-x", "schema": "vesta"}

        payload = {
            "query": "query GetLocalData($propertyId: ID!) {  home(property_id: $propertyId) {    local {      wildfire {        fsid        fire_factor_score        fire_factor_severity        fire_cumulative_30        fire_trend        fire_trend_paragraph        usfs_relative_risk        firststreet_url        fire_insurance_text        insurance_rates {          provider_logo          provider_url          providers        }      }    }  }}",
            "propertyId": property_id,
            "callfrom": "LDP",
            "nrQueryType": "WILDFIRE_RISK",
            "variables": {"propertyId": property_id},
            "isClient": True
        }

        response = requests.request(
            "POST", url, json=payload, headers=headers, params=querystring)
        return response.json()

    @staticmethod
    def get_value_estimates(
        property_id: str,
        start: datetime = datetime(datetime.now().year-3, 1, 1),
        end: datetime = datetime.now(),
        forecast_max: datetime = datetime(
            datetime.now().year, datetime.now().month + 6, 1),
        headers: Dict = defaults.HEADER
    ) -> Dict[str, Any]:
        """Gets the Realtor.com RealEstimate current and historic property value estimates

        Args:
            property_id (str): The Realtor.com property ID
            start (datetime, optional): The start time for the historic range. Defaults to datetime(datetime.now().year-3, 1, 1).
            end (datetime, optional): The end time for the historic range. Defaults to datetime.now().
            forecast_max (datetime, optional): The forecasting max. Defaults to datetime(datetime.now().year, datetime.now().month + 6, 1)
            headers (Dict, optional): The request headers. Incorrect headers may lead to invalid results. It is recommended to leave as default. Defaults to defaults.HEADER.

        Returns:
            Dict[str, Any]: The Relotr.com RealEstimate data
        """

        url = "https://www.realtor.com/api/v1/hulk"

        querystring = {"client_id": "rdc-x", "schema": "vesta"}

        payload = {
            "query": None,
            "queryLoader": {
                "appType": "FOR_SALE",
                "pageType": "LDP",
                "serviceType": "HOME_ESTIMATES"
            },
            "propertyId": property_id,
            "callfrom": "LDP",
            "nrQueryType": "HOME_ESTIMATES",
            "variables": {
                "propertyId": property_id,
                "historicalMin": start.strftime("%Y-%m-%d"),
                "historicalMax": end.strftime("%Y-%m-%d"),
                "forecastMax": forecast_max.strftime("%Y-%m-%d")
            },
            "isClient": True
        }

        response = requests.request(
            "POST", url, json=payload, headers=headers, params=querystring)
        return response.json()

    @staticmethod
    def get_nearby_home_values(property_id: str, headers: Dict = defaults.HEADER) -> List[Dict[str, Any]]:
        """Gets a list of nearby homes and their values

        Args:
            property_id (str): The Realtor.com property ID
            headers (Dict, optional): The request headers. Incorrect headers may lead to invalid results. It is recommended to leave as default. Defaults to defaults.HEADER.

        Returns:
            List[Dict[str, Any]]: The list of nearby homes and their values
        """

        url = "https://www.realtor.com/api/v1/hulk"

        querystring = {"client_id": "rdc-x", "schema": "vesta"}

        payload = {
            "query": None,
            "queryLoader": {
                "appType": "FOR_SALE",
                "pageType": "LDP",
                "serviceType": "NEARBY_HOME_VALUES"
            },
            "propertyId": property_id,
            "callfrom": "LDP",
            "nrQueryType": "NEARBY_HOME_VALUES",
            "variables": {"query": {
                "type": "nearby_homes",
                "property_id": property_id,
            }},
            "isClient": True,
            "isBot": "false"
        }

        response = requests.request(
            "POST", url, json=payload, headers=headers, params=querystring)
        return response.json()['data']['linked_homes']['results']

    @staticmethod
    def get_similar_homes(property_id: str, headers: Dict = defaults.HEADER) -> Dict[str, Any]:
        """Gets similar homes

        Args:
            property_id (str): The Realtor.com property ID
            headers (Dict, optional): The request headers. Incorrect headers may lead to invalid results. It is recommended to leave as default. Defaults to defaults.HEADER.

        Returns:
            List[Dict[str, Any]]: Similar
        """

        url = "https://www.realtor.com/api/v1/hulk"

        querystring = {"client_id": "rdc-x", "schema": "vesta"}

        payload = {
            "query": None,
            "propertyId": property_id,
            "callfrom": "LDP",
            "nrQueryType": "SIMILAR_HOMES",
            "ab_test_variation": None,
            "variables": {
                "propertyId": property_id,
                "relatedHomesQuery": {"type": "similar_homes"}
            },
            "isClient": True,
            "queryLoader": {
                "appType": "FOR_SALE",
                "pageType": "LDP",
                "serviceType": "SIMILAR_HOMES"
            },
            "isBot": "false"
        }

        response = requests.request(
            "POST", url, json=payload, headers=headers, params=querystring)
        return response.json()['data']['home']['related_homes']

    @staticmethod
    def get_homes_in_area_with_price(
        zip_code: str, price_min: int, price_max: int, limit: int = 15, offset: int = 0, headers: Dict = defaults.HEADER
    ) -> Dict[str, Any]:
        """Gets homes in a zip code that have a price within the specified range.

        Args:
            zip_code (str): The zip code to find homes in
            price_min (int): The minimum price
            price_max (int): The maximum price
            limit (int, optional): The limit of the number of results. Defaults to 15 (Realtor.com default)
            offset (int, optional): The offset on the results. If offset is 2, the first result will be the 3rd result. Defaults to 0.
            headers (Dict, optional): The request headers. Incorrect headers may lead to invalid results. It is recommended to leave as default. Defaults to defaults.HEADER.

        Returns:
            Dict[str, Any]: The homes in the zip code with price in range
        """

        url = "https://www.realtor.com/api/v1/hulk"

        querystring = {"client_id": "rdc-x", "schema": "vesta"}

        payload = {
            "callfrom": "LDP",
            "nrQueryType": "HOMES_AROUND_VALUE",
            "query": None,
            "queryLoader": {
                "appType": "FOR_SALE",
                "pageType": "LDP",
                "serviceType": "HOMES_AROUND_VALUE"
            },
            "variables": {
                "query": {
                    "primary": True,
                    "status": ["for_sale", "ready_to_build"],
                    "postal_code": zip_code,
                    "list_price": {
                        "min": price_min,
                        "max": price_max
                    }
                },
                "limit": limit,
                "offset": offset,
                "sort": {
                    "field": "list_date",
                    "direction": "desc"
                }
            }
        }

        response = requests.request(
            "POST", url, json=payload, headers=headers, params=querystring)
        return response.json()['data']['home_search']



import requests
from typing import List, Literal, Set, Dict, Any
from . import defaults
import re


class Query:

    def __init__(self) -> None:

        self.page: int = 1

        self.state: str | None = None
        self.region: str | None = None
        self.county: str | None = None
        self.city: str | None = None

        self.price_max: int | None = None
        self.price_min: int | None = None

        self.acres_max: int | None = None
        self.acres_min: int | None = None

        self.sqft_max: int | None = None
        self.sqft_min: int | None = None

        self.property_type: Set[Literal[
            "Commerical",
            "Farms and Ranches",
            "Homesite",
            "Horse",
            "House",
            "Hunting",
            "Lakefront",
            "Oceanfront",
            "Recreational",
            "Riverfront",
            "Timberland",
            "Undeveloped",
            "Waterfront",
        ]] | Literal[
            "Commerical",
            "Farms and Ranches",
            "Homesite",
            "Horse",
            "House",
            "Hunting",
            "Lakefront",
            "Oceanfront",
            "Recreational",
            "Riverfront",
            "Timberland",
            "Undeveloped",
            "Waterfront",
        ] | None = None

        self.beds_max: int | None = None
        self.beds_min: int | None = None

        self.baths_max: int | None = None
        self.baths_min: int | None = None

        self.activity: Literal[
            "boating",
            "fishing",
            "beach",
            "horseback riding",
            "rving",
            "canoeing/kayaking",
            "off-roading",
            "camping",
            "conservation",
            "aviation",
        ] | None = None

        self.available = True
        self.under_contract = False
        self.off_market = False
        self.sold = False

        self.sale_type: Literal['sale', 'auction', 'both'] = 'sale'

        self.owner_financing = False
        self.mineral_rights = False
        self.virtual_tour = False

        self.keywords: List[str] = []

    def create_url(self) -> str:
        """Creates the Landwatch GET request URL based on the query attributes

        Returns:
            str: The Landwatch GET request URL
        """

        url = defaults.ROOT_URL

        if self.state is None and self.property_type is None:
            url += "/land"

        if self.state is not None:
            url += Query_Helpers.get_link_for_state(self.state)

            if self.city is not None:
                url += Query_Helpers.get_link_for_city(self.city)
            elif self.county is not None:
                url += Query_Helpers.get_link_for_county(self.county)
            elif self.region is not None:
                url += Query_Helpers.get_link_for_region(self.region)

        if self.property_type is not None:
            url += Query_Helpers.get_link_for_property(self.property_type)

        if self.activity is not None:
            url += Query_Helpers.get_link_for_activity(self.activity)

        if self.price_min is not None or self.price_max is not None:
            url += Query_Helpers.get_link_for_price(
                self.price_min, self.price_max)

        if self.acres_min is not None or self.acres_max is not None:
            url += Query_Helpers.get_link_for_acres(
                self.acres_min, self.acres_max)

        if self.sqft_min is not None or self.sqft_max is not None:
            url += Query_Helpers.get_link_for_sqft(
                self.sqft_min, self.sqft_max)

        if self.beds_min is not None or self.beds_max is not None:
            url += Query_Helpers.get_link_for_beds(
                self.beds_min, self.beds_max)

        if self.baths_min is not None or self.baths_max is not None:
            url += Query_Helpers.get_link_for_baths(
                self.baths_min, self.baths_max)

        if self.keywords:
            url += Query_Helpers.get_link_for_keywords(self.keywords)

        if self.available:
            url += '/available'

        if self.under_contract:
            url += '/under-contract'

        if self.off_market:
            url += '/under-contract'

        if self.sold:
            url += '/sold'

        url += Query_Helpers.get_link_for_sale_type(self.sale_type)

        if self.mineral_rights:
            url += '/mineral-rights'

        if self.owner_financing:
            url += '/owner-financing'

        if self.virtual_tour:
            url += '/virtual-tour'

        if self.page > 1:
            url += f'/page-{self.page}'

        return url

    def set_page(self, page: int) -> 'Query':
        """Sets the page of the results

        Args:
            page (int): Page, must be 1 or greater

        Raises:
            ValueError: Page value less than 1

        Returns:
            Query: Returns self
        """

        if page < 1:
            raise ValueError(f'page must be 1 or greater, was {page}')

        self.page = page
        return self

    def set_state(self, state: str) -> 'Query':
        """Sets the state for the filter.

        Args:
            state (str): The state name

        Raises:
            ValueError: Invalid state name

        Returns:
            Query: Returns self
        """

        if state.lower() in Query_Helpers.state_dict:
            self.state = state
            return self
        else:
            raise ValueError(
                f'Invalid state name "{state}", valid names are {", ".join(Query_Helpers.state_dict.keys())}')

    def set_region(self, region: str) -> 'Query':
        """Sets the region for the filter. Note for this to work, must have a state specified and the region must be valid for said state

        Args:
            region (str): Region name

        Returns:
            Query: Returns self
        """

        self.region = region
        return self

    def set_county(self, county: str) -> 'Query':
        """Sets the county for the filter. Note for this to work, must have state specified and the county name must be valid for said state

        Args:
            county (str): County name

        Returns:
            Query: Returns self
        """

        self.county = county
        return self

    def set_city(self, city: str) -> 'Query':
        """Sets the city for the filter. Note for this to work, must have state specified and the city name must be valid for said state

        Args:
            city (str): City name

        Returns:
            Query: Returns self
        """

        self.city = city
        return self

    def set_price(self, price_min: int | None = None, price_max: int | None = None) -> 'Query':
        """Sets the price range for the filter

        Args:
            price_min (int | None, optional): Price min. If not specified, this end of the range will be left open. Defaults to None.
            price_max (int | None, optional): Price max. If not specified, this end of the range will be left open. Defaults to None.

        Returns:
            Query: Returns self
        """

        self.price_max = price_max
        self.price_min = price_min
        return self

    def set_acres(self, acres_min: int | None = None, acres_max: int | None = None) -> 'Query':
        """Sets the acres (lot area) for the filter

        Args:
            acres_min (int | None, optional): Acres min. If not specified, this end of the range will be left open. Defaults to None.
            acres_max (int | None, optional): Acres max. If not specified, this end of the range will be left open. Defaults to None.

        Returns:
            Query: Returns self
        """

        self.acres_max = acres_max
        self.acres_min = acres_min
        return self

    def set_sqft(self, sqft_min: int | None = None, sqft_max: int | None = None) -> 'Query':
        """Sets the sqft (interior area) for the filter

        Args:
            sqft_min (int | None, optional): Sqft min. If not specified, this end of the range will be left open. Defaults to None.
            sqft_max (int | None, optional): Sqft max. If not specified, this end of the range will be left open. Defaults to None.

        Returns:
            Query: Returns self
        """

        self.sqft_max = sqft_max
        self.sqft_min = sqft_min
        return self

    def set_property_type(self, property_type: Set[Literal[
            "Commerical",
            "Farms and Ranches",
            "Homesite",
            "Horse",
            "House",
            "Hunting",
            "Lakefront",
            "Oceanfront",
            "Recreational",
            "Riverfront",
            "Timberland",
            "Undeveloped",
            "Waterfront",
        ]] | Literal[
            "Commerical",
            "Farms and Ranches",
            "Homesite",
            "Horse",
            "House",
            "Hunting",
            "Lakefront",
            "Oceanfront",
            "Recreational",
            "Riverfront",
            "Timberland",
            "Undeveloped",
            "Waterfront",
    ]) -> 'Query':
        """Sets the property_type for the filter

        Args:
            property_type (Set[Literal[ &quot;Commerical&quot;, &quot;Farms and Ranches&quot;, &quot;Homesite&quot;, &quot;Horse&quot;, &quot;House&quot;, &quot;Hunting&quot;, &quot;Lakefront&quot;, &quot;Oceanfront&quot;, &quot;Recreational&quot;, &quot;Riverfront&quot;, &quot;Timberland&quot;, &quot;Undeveloped&quot;, &quot;Waterfront&quot;, ]] | Literal[ &quot;Commerical&quot;, &quot;Farms and Ranches&quot;, &quot;Homesite&quot;, &quot;Horse&quot;, &quot;House&quot;, &quot;Hunting&quot;, &quot;Lakefront&quot;, &quot;Oceanfront&quot;, &quot;Recreational&quot;, &quot;Riverfront&quot;, &quot;Timberland&quot;, &quot;Undeveloped&quot;, &quot;Waterfront&quot;, ]): The property type(s) either a single type given as a string or a set given as a set of strings.

        Raises:
            ValueError: Invalid property_type
            TypeError: Invalid property_type type given

        Returns:
            Query: Returns self
        """

        if isinstance(property_type, str):
            if property_type not in Query_Helpers.property_type_dict:
                raise ValueError(
                    f'Invalid property_type "{property_type}", valid property types are {", ".join(Query_Helpers.property_type_dict.keys())}')

            self.property_type = property_type
            return self

        elif isinstance(property_type, set):
            diff = property_type - set(Query_Helpers.property_type_dict.keys())
            if diff != set():
                raise ValueError(
                    f'Invalid property_types {", ".join(diff)}, valid property types are {", ".join(Query_Helpers.property_type_dict.keys())}'
                )

            self.property_type = property_type
            return self

        else:
            raise TypeError(
                f'property_type should either be a str or a Set[str], was {type(property_type)}')

    def set_beds(self, beds_min: int | None = None, beds_max: int | None = None) -> 'Query':
        """Sets the beds range for the filter

        Args:
            beds_min (int | None, optional): Beds min. If not specified, this end of the range will be left open. Defaults to None.
            beds_max (int | None, optional): Beds max. If not specified, this end of the range will be left open. Defaults to None.

        Returns:
            Query: Returns self
        """

        self.beds_max = beds_max
        self.beds_min = beds_min
        return self

    def set_baths(self, baths_min: int | None = None, baths_max: int | None = None) -> 'Query':
        """Sets the baths range for the filter

        Args:
            baths_min (int | None, optional): Baths min. If not specified, this end of the range will be left open. Defaults to None.
            baths_max (int | None, optional): Baths max. If not specified, this end of the range will be left open. Defaults to None.

        Returns:
            Query: Returns self
        """

        self.baths_max = baths_max
        self.baths_min = baths_min
        return self

    def set_activity(self, activity: Literal[
        "boating",
        "fishing",
        "beach",
        "horseback riding",
        "rving",
        "canoeing/kayaking",
        "off-roading",
        "camping",
        "conservation",
        "aviation",
    ]) -> 'Query':
        """Sets the activity for the filter

        Args:
            activity (Literal[ &quot;boating&quot;, &quot;fishing&quot;, &quot;beach&quot;, &quot;horseback riding&quot;, &quot;rving&quot;, &quot;canoeing): Activity

        Raises:
            ValueError: Invalid activity

        Returns:
            Query: Returns self
        """

        if activity not in Query_Helpers.activity_dict:
            raise ValueError(
                f'Invalid activity "{activity}", valid activities are {", ".join(Query_Helpers.activity_dict.keys())}')

        self.activity = activity
        return self

    def set_status(self, available=True, under_contract=False, off_market=False, sold=False) -> 'Query':
        """Sets the status of the property filter

        Args:
            available (bool, optional): If available. Defaults to True.
            under_contract (bool, optional): If under contract. Defaults to False.
            off_market (bool, optional): If off market. Defaults to False.
            sold (bool, optional): If sold. Defaults to False.

        Returns:
            Query: Returns self
        """

        self.available = available
        self.under_contract = under_contract
        self.off_market = off_market
        self.sold = sold
        return self

    def set_sale_type(self, sale_type: Literal['sale', 'auction', 'both']) -> 'Query':
        """Sets the sale type for the filter

        Args:
            sale_type (Literal[&#39;sale&#39;, &#39;auction&#39;, &#39;both&#39;]): The sale type

        Raises:
            ValueError: Invalid sales type

        Returns:
            Query: Returns self
        """

        valid = ['sale', 'auction', 'both']
        if sale_type not in valid:
            raise ValueError(
                f'Invalid sale type "{sale_type}", valid types are {", ".join(valid)}')

        self.sale_type = sale_type
        return self

    def set_owner_financing(self, owner_financing: bool) -> 'Query':
        """Sets if the owner is financing for the filter

        Args:
            owner_financing (bool): Owner financing

        Returns:
            Query: Returns self
        """

        self.owner_financing = owner_financing
        return self

    def set_mineral_rights(self, mineral_rights: bool) -> 'Query':
        """Sets if mineral rights are part of the property for the filter

        Args:
            mineral_rights (bool): Mineral rights

        Returns:
            Query: Returns self
        """

        self.mineral_rights = mineral_rights
        return self

    def set_virtual_tour(self, virtual_tour: bool) -> 'Query':
        """Sets if there is a virtual tour for the property for the filter

        Args:
            virtual_tour (bool): Virtual tour

        Returns:
            Query: Returns self
        """

        self.virtual_tour = virtual_tour
        return self

    def set_keywords(self, keywords: List[str]) -> 'Query':
        """Sets the keywords for the filter

        Args:
            keywords (List[str]): List of keywords. These must be alphanumeric with no spaces.

        Raises:
            ValueError: Invalid format for keyword

        Returns:
            Query: Returns self
        """

        for keyword in keywords:
            if not re.match(r'[A-Za-z0-9]+', keyword):
                raise ValueError(
                    f'Keywords must be alphanumeric with no spaces or symbols, keyword "{keyword}" was not')

        self.keywords = keywords

    def set_all(self, **kwargs) -> 'Query':
        """Meta setter function that can set all attributes

        Returns:
            Query: Returns self
        """

        if x := kwargs.get('page'):
            self.set_page(x)

        if x := kwargs.get('state'):
            self.set_state(x)

        if x := kwargs.get('region'):
            self.set_region(x)

        if x := kwargs.get('county'):
            self.set_county(x)

        if x := kwargs.get('city'):
            self.set_city(x)

        self.set_price(kwargs.get('price_min'), kwargs.get('price_max'))
        self.set_acres(kwargs.get('acres_min'), kwargs.get('acres_max'))
        self.set_sqft(kwargs.get('sqft_min'), kwargs.get('sqft_max'))

        if x := kwargs.get('property_type'):
            self.set_property_type(x)

        self.set_beds(kwargs.get('beds_min'), kwargs.get('beds_max'))
        self.set_baths(kwargs.get('baths_min'), kwargs.get('baths_max'))

        if x := kwargs.get('activity'):
            self.set_activity(x)

        self.set_status(
            kwargs.get('available', True),
            kwargs.get('under_contract', False),
            kwargs.get('off_market', False),
            kwargs.get('sold', False)
        )

        if x := kwargs.get('sale_type'):
            self.set_sale_type(x)

        if x := kwargs.get('owner_financing'):
            self.set_owner_financing(x)

        if x := kwargs.get('mineral_rights'):
            self.set_mineral_rights(x)

        if x := kwargs.get('virtual_tour'):
            self.set_virtual_tour(x)

        if x := kwargs.get('keywords'):
            self.set_keywords(x)

        return self

    def get_response(self, headers: Dict = defaults.HEADER) -> requests.Response:
        """Gets the GET request response for the Landwatch API given filter parameters

        Args:
            headers (Dict, optional): The request headers. Incorrect headers may lead to the Landwatch PUBLIC api blocking the request. Defaults to defaults.HEADER.

        Returns:
            requests.Response: GET request response object
        """

        return requests.get(
            self.create_url(),
            headers=headers
        )

    def get_results(self, returns: Literal["full", "results"] = "results", headers: Dict = defaults.HEADER) -> Dict[str, Any]:
        """Gets the results from the GET request to the PUBLIC Landwatch API given filter parameters.

        Args:
            returns (Literal[&quot;full&quot;, &quot;results&quot;], optional): What it should return, either the 'full' json content of the response or 'results' - the part that pertains to the search results. Defaults to "results".
            headers (Dict, optional): The request headers. Incorrect headers may lead to the Landwatch PUBLIC api blocking the request. Defaults to defaults.HEADER.

        Raises:
            ValueError: Invalid returns parameter value

        Returns:
            Dict[str, Any]: The response data
        """

        data = self.get_response(headers).json()

        if returns == 'full':
            return data
        elif returns == 'results':
            return data['searchResults']
        else:
            raise ValueError(
                f'Invalid returns value "{returns}", valid options are "full" and "results"')

    def get_filter_options(self, headers: Dict = defaults.HEADER) -> Dict[str, Any]:
        """Gets the potential filter options, such as possible states (or regions if state was set).

        Args:
            headers (Dict, optional): The request headers. Incorrect headers may lead to the Landwatch PUBLIC api blocking the request. Defaults to defaults.HEADER.

        Returns:
            Dict[str, Any]: Dictionary of filter options
        """

        data = self.get_results('full')
        options = {}
        for fs in data['filterSections']:
            options[fs['section']] = fs['filterLinks']

        return options


class Query_Helpers:

    state_dict = {
        "texas": "/texas-land-for-sale",
        "florida": "/florida-land-for-sale",
        "georgia": "/georgia-land-for-sale",
        "north carolina": "/north-carolina-land-for-sale",
        "california": "/california-land-for-sale",
        "tennessee": "/tennessee-land-for-sale",
        "south carolina": "/south-carolina-land-for-sale",
        "michigan": "/michigan-land-for-sale",
        "arizona": "/arizona-land-for-sale",
        "new york": "/new-york-land-for-sale",
        "arkansas": "/arkansas-land-for-sale",
        "virginia": "/virginia-land-for-sale",
        "kentucky": "/kentucky-land-for-sale",
        "missouri": "/missouri-land-for-sale",
        "oklahoma": "/oklahoma-land-for-sale",
        "pennsylvania": "/pennsylvania-land-for-sale",
        "indiana": "/indiana-land-for-sale",
        "wisconsin": "/wisconsin-land-for-sale",
        "alabama": "/alabama-land-for-sale",
        "ohio": "/ohio-land-for-sale",
        "illinois": "/illinois-land-for-sale",
        "oregon": "/oregon-land-for-sale",
        "mississippi": "/mississippi-land-for-sale",
        "colorado": "/colorado-land-for-sale",
        "idaho": "/idaho-land-for-sale",
        "new mexico": "/new-mexico-land-for-sale",
        "minnesota": "/minnesota-land-for-sale",
        "iowa": "/iowa-land-for-sale",
        "louisiana": "/louisiana-land-for-sale",
        "nevada": "/nevada-land-for-sale",
        "utah": "/utah-land-for-sale",
        "maryland": "/maryland-land-for-sale",
        "montana": "/montana-land-for-sale",
        "new jersey": "/new-jersey-land-for-sale",
        "washington": "/washington-land-for-sale",
        "west virginia": "/west-virginia-land-for-sale",
        "kansas": "/kansas-land-for-sale",
        "maine": "/maine-land-for-sale",
        "connecticut": "/connecticut-land-for-sale",
        "new hampshire": "/new-hampshire-land-for-sale",
        "massachusetts": "/massachusetts-land-for-sale",
        "wyoming": "/wyoming-land-for-sale",
        "south dakota": "/south-dakota-land-for-sale",
        "delaware": "/delaware-land-for-sale",
        "vermont": "/vermont-land-for-sale",
        "hawaii": "/hawaii-land-for-sale",
        "alaska": "/alaska-land-for-sale",
        "nebraska": "/nebraska-land-for-sale",
        "north dakota": "/north-dakota-land-for-sale",
        "rhode island": "/rhode-island-land-for-sale",
        "district of columbia": "/district-of-columbia-land-for-sale",
    }

    activity_dict = {
        "boating": "/boating-activity",
        "fishing": "/fishing-activity",
        "beach": "/beach-activity",
        "horseback riding": "/horseback-riding-activity",
        "rving": "/rving-activity",
        "canoeing/kayaking": "/canoeing-kayaking-activity",
        "off-roading": "/off-roading-activity",
        "camping": "/camping-activity",
        "conservation": "/conservation-activity",
        "aviation": "/aviation-activity",
    }

    property_type_dict = {
        "house": "/homes",
        "undeveloped": "/undeveloped-land",
        "homesite": "/homesites",
        "waterfront": "/waterfront-property",
        "lakefront": "/lakefront-property",
        "commercial": "/commercial-property",
        "recreational": "/recreational-property",
        "farms and ranches": "/farms-ranches",
        "hunting": "/hunting-property",
        "timberland": "/timberland-property",
        "horse": "/horse-property",
        "riverfront": "/riverfront-property",
        "oceanfront": "/oceanfront-property",
    }

    property_type_id_dict = {
        "house": 8192,
        "undeveloped": 32,
        "homesite": 4096,
        "waterfront": 3584,
        "lakefront": 512,
        "commercial": 64,
        "recreational": 4,
        "farms and ranches": 3,
        "hunting": 128,
        "timberland": 16,
        "horse": 256,
        "riverfront": 2048,
        "oceanfront": 1024,
    }

    @classmethod
    def get_link_for_state(cls, state: str) -> str:
        """Gets the relative link for the filtering for the state

        Args:
            state (str): The state name

        Raises:
            ValueError: Invalid state name

        Returns:
            str: relative link
        """

        try:
            return cls.state_dict[state.lower()]
        except KeyError:
            raise ValueError(
                f'{state} is not a valid state name, valid state names are {", ".join(cls.state_dict.keys())}')

    @classmethod
    def get_link_for_property(cls, property_type: Set[str] | str) -> str:
        """Gets the link for the property type filter

        Args:
            property_type (Set[str] | str): Either a single property type as a string or a set of property type strings

        Raises:
            ValueError: Not a valid property type
            TypeError: property_type is neither a str or a set

        Returns:
            str: _description_
        """

        if isinstance(property_type, str):
            try:
                return cls.property_type_dict[property_type.lower()]
            except KeyError:
                raise ValueError(
                    f'{property_type} is not a valid propety type, valid propety types are {", ".join(cls.property_type_dict.keys())}')

        elif isinstance(property_type, set):

            id_sum = 0

            for pt in property_type:
                try:
                    id_sum += cls.property_type_id_dict[pt]
                except KeyError:
                    raise ValueError(
                        f'{pt} is not a valid propety type, valid propety types are {", ".join(cls.property_type_dict.keys())}')

            if id_sum > 0:
                return f'/prop-types-{id_sum}'
            else:
                return ''

        else:
            raise TypeError(
                f'property_type parameter is an invalid type {type(property_type)}. Must be either a str or a Set[str]')

    @staticmethod
    def get_link_for_region(region: str) -> str:
        """Formats the region name as needed by the Landwatch api, all lower case, spaces replaced with -, '-region' added at the end.

        Args:
            region (str): Region name

        Returns:
            str: Region name formatted ('/region-name-region')
        """
        region = region.lower().replace(' ', '-')
        return f'/{region}-region'

    @staticmethod
    def get_link_for_county(county: str) -> str:
        """Formats the county name as needed by the Landwatch api, all lower case, spaces replaced with -, '-county' added at the end.

        Args:
            county (str): County name

        Returns:
            str: County name formatted ('/county-name-county')
        """
        county = county.lower().replace(' ', '-')
        return f'/{county}-county'

    @staticmethod
    def get_link_for_city(city: str) -> str:
        """Formats the city name as needed by the Landwatch api, all lower case, spaces replaced with -

        Args:
            city (str): The city name

        Returns:
            str: City name formatted for landwatch ('/city-name')
        """

        city = city.lower().replace(' ', '-')
        return f'/{city}'

    @staticmethod
    def get_link_for_price(price_min: int | None = None, price_max: int | None = None) -> str:
        """Gets the link for the price filter.

        Args:
            price_min (int | None, optional): Min price in USD. Defaults to None.
            price_max (int | None, optional): Max price in USD. Defaults to None.

        Returns:
            str: the link
        """

        if price_max is not None and price_min is not None:
            return f'/price-{price_min}-{price_max}'
        elif price_max is not None:
            return f'/price-under-{price_max}'
        elif price_min is not None:
            return f'/price-over-{price_min}'
        else:
            return ''

    @staticmethod
    def get_link_for_acres(acres_min: int | None = None, acres_max: int | None = None) -> str:
        """Gets the link for the lot size filter

        Args:
            acres_min (int | None, optional): Lot size area in Acres. Defaults to None.
            acres_max (int | None, optional): Lot size area in acres. Defaults to None.

        Returns:
            str: The link for the size filter
        """
        if acres_max is not None and acres_min is not None:
            return f'/acres-{acres_min}-{acres_max}'
        elif acres_max is not None:
            return f'/acres-under-{acres_max}'
        elif acres_min is not None:
            return f'/acres-over-{acres_min}'
        else:
            return ''

    @staticmethod
    def get_link_for_sqft(sqft_min: int | None = None, sqft_max: int | None = None) -> str:
        """Gets the link for the lot size filter

        Args:
            sqft_min (int | None, optional): Interior area min in sqft. Defaults to None.
            sqft_max (int | None, optional): Interior area max in sqft. Defaults to None.

        Returns:
            str: The link for the size filter
        """
        if sqft_max is not None and sqft_min is not None:
            return f'/sqft-{sqft_min}-{sqft_max}'
        elif sqft_max is not None:
            return f'/sqft-under-{sqft_max}'
        elif sqft_min is not None:
            return f'/sqft-over-{sqft_min}'
        else:
            return ''

    @staticmethod
    def get_link_for_beds(beds_min: int | None = None, beds_max: int | None = None) -> str:
        """Gets the link for the number of beds filter

        Args:
            beds_min (int | None, optional): Min number of beds. Defaults to None.
            beds_max (int | None, optional): Max number of beds. Defaults to None.

        Returns:
            str: The link for the number of beds filter
        """
        if beds_max is not None and beds_min is not None:
            return f'/beds-{beds_min}-{beds_max}'
        elif beds_max is not None:
            return f'/beds-under-{beds_max}'
        elif beds_min is not None:
            return f'/beds-over-{beds_min}'
        else:
            return ''

    @staticmethod
    def get_link_for_baths(baths_min: int | None = None, baths_max: int | None = None) -> str:
        """Gets the link for the number of baths filter

        Args:
            baths_min (int | None, optional): Min number of baths. Defaults to None.
            baths_max (int | None, optional): Max number of baths. Defaults to None.

        Returns:
            str: The link for the number of baths filter
        """
        if baths_max is not None and baths_min is not None:
            return f'/baths-{baths_min}-{baths_max}'
        elif baths_max is not None:
            return f'/baths-under-{baths_max}'
        elif baths_min is not None:
            return f'/baths-over-{baths_min}'
        else:
            return ''

    @classmethod
    def get_link_for_activity(cls, activity: str) -> str:
        """Gets the relative link for filtering based on the specified activity

        Args:
            activity (str): activity

        Raises:
            ValueError: Invalid activity name

        Returns:
            str: relative link
        """
        try:
            return cls.activity_dict[activity.lower()]
        except KeyError:
            raise ValueError(
                f'{activity} is not a valid activity name, valid activity names are {", ".join(cls.activity_dict.keys())}')

    @staticmethod
    def get_link_for_sale_type(sale_type: Literal['sale', 'auction', 'both']) -> str:
        """Gets the link for the specified sale type

        Args:
            sale_type (Literal[&#39;sale&#39;, &#39;auction&#39;, &#39;both&#39;]): The sale type

        Raises:
            ValueError: Invalid sale type given

        Returns:
            str: the link for the specified sale type
        """

        if sale_type == 'both':
            return ''
        elif sale_type == 'sale':
            return '/for-sale'
        elif sale_type == 'auction':
            return '/auctions'
        else:
            raise ValueError('Invalid sale type')

    @staticmethod
    def get_link_for_keywords(keywords: List[str]) -> str:
        """Gets the link for filtering based on specified keywords

        Args:
            keywords (List[str]): Keywords. Note that keywords can only be alphanumeric, no spaces or symbols. Keywords are case insensitive

        Raises:
            ValueError: Invalid keyword format

        Returns:
            str: Keyword filter link
        """

        link = '/keyword'

        for keyword in keywords:
            if re.match(r'[A-Za-z0-9]+', keyword):
                link += f'-{keyword.lower()}'
            else:
                raise ValueError(
                    f'Keywords must be alphanumeric with no spaces or symbols, keyword "{keyword}" was not')

        return link



import json
import requests
from bs4 import BeautifulSoup
from numbers import Number
from typing import Dict, Any, Tuple, List
from .. import defaults


class Details_Page:
    """Static class which contains methods that are more or less generalizable to detail page parsing of Sales, Rental Homes, and Rental Apartments
    """

    @staticmethod
    def get_page(url: str, headers: Dict = defaults.HEADER) -> requests.Response:
        """Submits a GET request to the detail URL to get the page.

        Args:
            url (str): The detail URL. Note that invalid URL will lead to unexpected results and errors
            headers (Dict, optional): The request headers. Incorrect headers may lead to invalid results. It is recommended to leave as default. Defaults to defaults.HEADER.

        Returns:
            requests.Response: The page GET response
        """
        return requests.get(url, headers=headers)

    @staticmethod
    def make_soup(page: requests.Response) -> BeautifulSoup:
        """This is a very simple function that has barely a reason to exist. Just takes the response content and has it parsed by Beautiful Soup via its html parser.

        Args:
            page (requests.Response): The GET response from a valid GET requests on the detail URL page

        Returns:
            BeautifulSoup: Page HTML parsed as a BeautifulSoup Object
        """

        soup = BeautifulSoup(page.content, "html.parser")

        return soup

    @staticmethod
    def get_walk_and_bike_score(zpid: str) -> Dict[str, Any]:
        """Gets the walk and bike details that appears on the Zillow details page

        Args:
            zpid (str): The Zillow property ID.

        Returns:
            Dict[str, Any]: Walk and bike score response dictionary
        """

        url = "https://www.zillow.com/graphql"

        payload = {
            "clientVersion": "home-details/6.1.1569.master.099cd8a",
            "operationName": "WalkTransitAndBikeScoreQuery",
            "query": "query WalkTransitAndBikeScoreQuery($zpid: ID!) {\n  property(zpid: $zpid) {\n    id\n    walkScore {\n      walkscore\n      description\n      ws_link\n    }\n    transitScore {\n      transit_score\n      description\n      ws_link\n    }\n    bikeScore {\n      bikescore\n      description\n    }\n  }\n}\n",
            "variables": {"zpid": zpid}
        }

        return requests.request(
            "POST", url, json=payload, headers=defaults.HEADER
        ).json()['data']

    @classmethod
    def parse_lot_size(cls, lot_size: str) -> Number:
        """Parses lot size string to sqft. Assumes that lot size will follow format of 'x Acres' or 'x.x Acres'.

        Args:
            lot_size (str): Zillow lot size string.

        Returns:
            Number: Lot size in sqft
        """
        if not lot_size:
            return None

        lot_size = lot_size.lower()

        if 'acres' in lot_size:
            acres = float(lot_size.replace(
                "acres", "").replace(",", "").strip())
            return cls.calculate_acres_to_sqft(acres)
        elif "sqft" in lot_size:
            sqft = float(lot_size.replace("sqft", "").replace(",", "").strip())
            return sqft
        else:
            raise ValueError(
                f"Expected to find 'acres' or 'sqft' but found neither - {lot_size}")

    @staticmethod
    def calculate_acres_to_sqft(acres: Number) -> Number:
        """Converts acres to sqft. 

        Conversion factor from https://www.metric-conversions.org/area/acres-to-square-feet.htm

        Args:
            acres (Number): Acres

        Returns:
            Number: sqft
        """
        return acres * 43560


class NextJS_Detail_Page(Details_Page):

    @staticmethod
    def get_next_data(soup: BeautifulSoup) -> Dict[Any, Any]:
        """Extracts and parses the NEXT_DATA cache from the page soup.

        Args:
            soup (BeautifulSoup): The html page soup object

        Returns:
            Dict[Any, Any]: The api preload cache dictionary
        """

        ndata = soup.find("script", id="__NEXT_DATA__").text
        ndata = json.loads(ndata)

        return ndata

    @staticmethod
    def get_initial_data_and_redux_state(ndata: Dict[Any, Any]) -> Tuple[Dict[Any, Any], Dict[Any, Any]]:
        """Gets the initial Data and Redux State from the ndata (next data) dictionary.

        Args:
            ndata (Dict[Any, Any]): The ndata dictionary obtained via get_next_data

        Returns:
            Tuple[Dict[Any, Any], Dict[Any, Any]]: initialData dictionary, initialReduxState dictionary
        """

        props = ndata['props']
        return props['initialData'], props['initialReduxState']


class Preload_Detail_Page(Details_Page):

    def __init__(self, url: str) -> None:
        """This initializes the Preload_Detail_Page object, which call a GET request on the Zillow detail URL. 

        Args:
            url (str): The Zillow Property Sale details URL.
        """

        self.url = url

        # get soup
        self.soup = self.make_soup(self.get_page(self.url))

        # get api preload
        self.preload = self.get_api_preload(self.soup)
        self.zpid: int = self.preload['zpid']

        # get caches
        self.variant_cache, self.full_cache = self.get_variant_and_full_from_preload(
            self.preload)

        # quick access values
        self.property: Dict[Any, Any] = self.full_cache['property']

        self.home_type: str = self.property['homeType']
        self.year_built: int = self.property['yearBuilt']

        self.price: int = self.property['price']
        self.zestimate: int = self.property['zestimate']
        self.rental_zestimate: int = self.property['rentZestimate']
        self.tax_history: List[Dict[str, Number]] = self.property['taxHistory']
        self.price_history: List[Dict[str, Any]
                                 ] = self.property['priceHistory']
        self.currency: str = self.property['currency']

        self.status: str = self.property['homeStatus']
        self.days_on_zillow: int = self.property['daysOnZillow']
        self.views: int = self.property['pageViewCount']
        self.saves: int = self.property['favoriteCount']

        self.tags: List[str] = self.get_tags()
        self.description: str = self.property['description']

        self.address: Dict[str, Any] = self.property['address']
        self.street_address: str = self.address['streetAddress']
        self.city: str = self.address['city']
        self.state: str = self.address['state']
        self.zip: str = self.address['zipcode']
        self.latitude: float = self.property['latitude']
        self.longitutde: float = self.property['longitude']

        self.bedrooms: Number = self.property['bedrooms']
        self.bathrooms: Number = self.property['bathrooms']

        self.interior_sqft: Number = self.property['livingArea']

        self.appliances: List[str] = self.property['resoFacts']['appliances']
        self.cooling: List[str] = self.property['resoFacts']['cooling']
        self.heating: List[str] = self.property['resoFacts']['heating']
        self.community_features: List[str] = self.property['resoFacts']['communityFeatures']
        self.fireplaces: Number = self.property['resoFacts']['fireplaces']
        self.garage: bool = self.property['resoFacts']['hasGarage']
        self.interior_features: List[str] = self.property['resoFacts']['interiorFeatures']

        self.attic: str | None = self.property['resoFacts']['attic']
        self.basement: str | None = self.property['resoFacts']['basement']

        self.hoa_fee: Number | None = self.property['resoFacts']['hoaFee']

        self.levels: str | None = self.property['resoFacts']['levels']

        self.parking: List[str] | None = self.property['resoFacts']['parkingFeatures']
        self.lot_features: List[str] | None = self.property['resoFacts']['lotFeatures']
        self.lot_size: str | None = self.property['resoFacts']['lotSize']
        self.lot_size_dimensions: str = self.property['resoFacts'].get(
            'lotSizeDimensions', None)
        self.lot_sqft: Number = self.parse_lot_size(self.lot_size)

        self.sewer: List[str] | None = self.property['resoFacts']['sewer']
        self.water_source: List[str] | None = self.property['resoFacts']['waterSource']

        self.attribution: Dict[str, Any] = self.property['attributionInfo']

        self.schools: List[Dict[str, Any]] = self.property['schools']
        self.similar: List[Dict[str, Any]] = self.property['comps']
        self.nearby: List[Dict[str, Any]] = self.property['nearbyHomes']

    @staticmethod
    def get_api_preload(soup: BeautifulSoup) -> Dict[Any, Any]:
        """Extracts and parses the API preload cache from the page soup.

        Args:
            soup (BeautifulSoup): The html page soup object

        Returns:
            Dict[Any, Any]: The api preload cache dictionary
        """

        preload = soup.find("script", id="hdpApolloPreloadedData").text
        preload = json.loads(preload)
        preload['apiCache'] = json.loads(preload['apiCache'])
        return preload

    @staticmethod
    def get_variant_and_full_from_preload(preload: Dict[Any, Any]) -> Tuple[Dict[Any, Any], Dict[Any, Any]]:
        """Gets the variant and full API caches from the preload dictionary.

        Args:
            preload (Dict[Any, Any]): The preload dictionary obtained via get_api_preload

        Raises:
            KeyError: Unexpected key within apiCache

        Returns:
            Tuple[Dict[Any, Any], Dict[Any, Any]]: variant api cache dictionary, full api cache dictionary
        """

        for k in preload['apiCache'].keys():
            if 'Variant' in k:
                var_key = k
            elif 'Full' in k:
                full_key = k
            else:
                raise KeyError(
                    "Unexpected key in apiCache dictionary of preload dictionary")

        return preload['apiCache'][var_key], preload['apiCache'][full_key]

    def get_at_a_glance(self) -> Dict[str, Any]:
        """Gets the Zillow at a glance facts.

        Returns:
            Dict[str, Any]: Returns a dictionary of the at a glance facts.
        """

        at_a_glance_dict = self.property['resoFacts']['atAGlanceFacts']

        if at_a_glance_dict:
            glance = {}
            for pair in at_a_glance_dict:
                glance[pair['factLabel']] = pair['factValue']

            return glance
        else:
            return {}

    def get_tags(self) -> List[str]:
        """Gets the taglines on the page

        Returns:
            List[str]: List of taglines
        """
        homeinsights = self.property.get('homeInsights', [{}])
        if not homeinsights:
            return []

        return list(
            {
                tag for tag_model in homeinsights[0]
                .get('insights', {}) for tag in tag_model.get('phrases', [])
            }
        )

    def get_facts_and_features(self) -> Dict[str, Any]:
        """Gets the Zillow facts and features section of the webpage as a dictionary

        Returns:
            Dict[str, Any]: Facts and features dictionary
        """

        return {tag.h5.text: {
            stag.h6.text: [
                li.text for li in stag.ul.find_all('li')
            ] for stag in tag.find_all("div")
        } for tag in self.soup.find_all("div", "jCOrgb")}



from typing import Dict, Any


class Listing_Details:

    def __init__(self, property_result: Dict[str, Any]) -> None:

        self.check_property_results_keys(property_result)

        self.price: int = property_result['price']

        self.acres: int = property_result['acres']
        self.sqft: int = property_result['homesqft']

        self.address: str = property_result['address']
        self.city: str = property_result['city']
        self.county: str = property_result['county']
        self.state: str = property_result['state']
        self.state_code: str = property_result['stateCode']
        self.zip: str = property_result['zip']
        self.lake: str = property_result['lake']
        self.longitude: float = property_result['longitude']
        self.latitude: float = property_result['latitude']

        self.auction_date: str = property_result['auctionDate']

        self.description: str = property_result['description']
        self.has_house: bool = property_result['hasHouse']
        self.has_video: bool = property_result['hasVideo']
        self.has_virtual_tour: bool = property_result['hasVirtualTour']
        self.labels: str = property_result['propertyTypesLabel']

        self.baths: float = property_result['baths']
        self.half_baths: float = property_result['halfBaths']
        self.beds: float = property_result['beds']

        self.broker_company: str = property_result['brokerCompany']
        self.broker_name: str = property_result['brokerName']

        self.url: str = f"https://www.landwatch.com{property_result['canonicalUrl']}"

    @staticmethod
    def check_property_results_keys(property_result: Dict[str, Any]) -> None:
        """Checks for the expected keys in a Landwatch property_result dictionary/json structure

        Args:
            property_result (Dict[str, Any]): Landwatch property_result dictionary/json

        Raises:
            KeyError: Missing keys
        """

        EXPECTED_KEYS = {
            'accountId', 'acres', 'acresDisplay', 'adTargetingCountyId', 'address', 'auctionDate', 'baths', 'bathsDisplay', 'bedsDisplay', 'beds', 'brokerCompany', 'brokerName', 'canonicalUrl', 'city', 'cityID', 'companyLogoDocumentId', 'county', 'countyId', 'countyLabel', 'description', 'encodedBoundaryPoints', 'externalSourceId', 'halfBaths', 'halfBathsDisplay', 'hasHouse', 'hasVideo', 'hasVirtualTour', 'homesqft', 'homesqftDisplay', 'imageCount', 'imageAltTextDisplay',
            'id', 'isALC', 'isDiamond', 'isFirstFreeListing', 'isGold', 'isHeadlineAd', 'isLiked', 'isPlatinum', 'isShowcase', 'lake', 'latitude', 'listHubListingKey', 'listingLevel', 'listingLevelTitle', 'longitude', 'partnerId', 'portraitDocumentId', 'price', 'priceChange', 'priceDisplay', 'propertyTypes', 'propertyTypesLabel', 'schemaData', 'shortPrice', 'siteListingId', 'state', 'stateAbbreviation', 'stateCode', 'stateId', 'status', 'thumbnailDocumentId', 'title', 'types', 'zip'
        }

        if EXPECTED_KEYS == set(property_result.keys()):
            return  # all expected keys are there

        missing = EXPECTED_KEYS - set(property_result.keys())

        raise KeyError(
            f'property_result is missing the following keys: {", ".join(missing)}')



from setuptools import setup, find_packages
from src.realty.version import __version__

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
license_text = (this_directory / "LICENSE").read_text()

setup(
    name='py-realty',
    version=__version__,
    description='Python real estate scraping library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license=license_text,
    author='Jakub Polanowski',
    url='https://github.com/JakubPolanowski/PY-Realty',
    python_requires=">=3.10, <4",
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=[
        "beautifulsoup4>=4.11.1",
        "numpy>=1.24.0",
        "requests>=2.28.1",
    ],
    setup_requires=['pytest-runner', 'flake8'],
    tests_require=['pytest'],
    project_urls={
        # "Documentation": "#TODO",
        "Source": "https://github.com/JakubPolanowski/PY-Realty",
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries",
    ],
)



URL = "https://www.zillow.com/search/GetSearchPageState.htm"
# The URL for the API request

HEADER = {  # Headers to be passed in the request (required to get results)
    "authority": "www.zillow.com",
    "accept": "*/*",
    "accept-language": "en-US,en;q=0.9",
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "cors",
    "sec-fetch-site": "same-origin",
    "sec-gpc": "1",
    "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36"
}

GRAPHQL_HEADER = {
    "authority": "www.zillow.com",
    "accept": "*/*",
    "accept-language": "en-US,en;q=0.7",
    "client-id": "vertical-living",
    "content-type": "application/json",
    "origin": "https://www.zillow.com",
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "cors",
    "sec-fetch-site": "same-origin",
    "sec-gpc": "1",
    "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36"
}



import json
import requests
from typing import Dict, Any, List, Literal, Union, Set
from . import defaults


class Query:

    def __init__(self):
        """Initializes defaults. Note that defaults are based on assumptions of how the Zillow API works, and are not necessarily optimal. 
        """

        # TODO further experiment to reduce unnecessary params

        self.sub_parms = {
            "pagination": {"currentPage": 1},
            "isMapVisible": "true",
            "isListVisible": "true",
        }

        self.wants = {
            "cat1": ["listResults", "mapResults"],
            "cat2": ["total"]
        }

    def set_page(self, current_page: int) -> 'Query':
        """Sets current page of paginated search results

        Args:
            current_page (int): Page of results to lookup

        Returns:
            Query: Returns self
        """
        self.sub_parms["pagination"] = {"currentPage": current_page}
        return self

    def set_search_term(self, term: str) -> 'Query':
        """Sets the query search term.

        Args:
            term (str): The search term for the query

        Returns:
            Query: Returns self
        """

        self.sub_parms["usersSearchTerm"] = term
        return self

    def set_map_bounds(self, west: float, east: float, south: float, north: float) -> 'Query':
        """Sets the map bounds for the query

        Args:
            west (float): West bound, longitude
            east (float): East bound, longitude
            south (float): South bound, latitude
            north (float): North bound, latitude 

        Returns:
            Query: Returns self
        """
        self.sub_parms["mapBounds"] = {
            "west": -85.58309022900734,
            "east": -85.18483583447609,
            "south": 34.73748469750508,
            "north": 35.40678434232889,
        }
        return self

    def set_region(self, region_id: int, region_type: int) -> 'Query':
        """Sets the region. Note that the region ID and Type appear to be internal to Zillow, therefore this is a difficult query param to use

        Args:
            region_id (int): Zillow Region ID value
            region_type (int): Zillow Region Type value

        Returns:
            Query: Returns self
        """
        self.sub_parms["regionSelection"] = [
            {"regionId": region_id, "regionType": region_type}]
        return self

    def set_map_visable(self, map_visible: bool) -> 'Query':
        # TODO determine what impact this has
        self.sub_parms["isMapVisible"] = map_visible
        return self

    def set_filter(self, filter_dict: Dict[str, Any]) -> 'Query':
        """Sets the filter state. This works by expanding the filter_dict from {key: value} to {key: {'value': value}} which is the format that the API accepts.

        Args:
            filter_dict (Dict[str, Any]): Filter parameters given as a dictionary of keys and values

        Returns:
            Query: Returns self
        """

        self.sub_parms["filterState"] = filter_dict
        return self

    def set_filter_preset(
            self,
            for_sale: bool = True,
            all_homes: bool = True,
            sale_options: Set[Literal[
                "ForSaleByAgent",
                "ForSaleByOwner",
                "NewConstruction",
                "ComingSoon",
                "Auction",
                "ForSaleForeclosure",
                "RecentlySold"
            ]] = {"ForSaleByAgent", "ForSaleByOwner", "NewConstruction", "ComingSoon", "Auction", "ForSaleForeclosure"},
            sold_in_last_x_days: Union[None, int] = None,
            price_max: Union[None, int] = None,
            price_min: Union[None, int] = None,
            monthly_payment_max: Union[None, int] = None,
            monthly_payment_min: Union[None, int] = None,
            beds: Union[None, int] = None,
            baths: Union[None, int] = None,
            home_type: Set[Literal[
                "SingleFamily",
                "Townhouse",
                "MultiFamily",
                "Condo",
                "LotLand",
                "Apartment",
                "Manufactured",
                "ApartmentOrCondo",
            ]] = {"SingleFamily", "Townhouse", "MultiFamily", "Condo", "LotLand", "Apartment", "Manufactured", "ApartmentOrCondo"},
            hoa: Union[None, int] = None,
            parking_spots: Union[None, int] = None,
            features: Set[Literal[
                "Garage",
                "BasementFinished",
                "BasementUnfinished",
                "SingleStory",
                "AgeRestricted55Plus",
                "AirConditioning",
                "Pool",
                "Waterfront",
                "CityView",
                "ParkView",
                "MountainView",
                "WaterView",
            ]] = set(),
            house_sqft_max: Union[None, int] = None,
            house_sqft_min: Union[None, int] = None,
            lot_sqft_max: Union[None, int] = None,
            lot_sqft_min: Union[None, int] = None,
            year_built_start: Union[None, int] = None,
            year_built_end: Union[None, int] = None,
            keywords: Set[str] = set(),
    ) -> 'Query':
        """This sets the filter_state based on the parameters given. This is used for the same purpose as set_filter but this is easier to use.

        Args:
            for_sale (bool, optional): If should search for sales, if set to False will search for rentals. Note that sale_options do not apply to rentals and will be ignored. Defaults to True.
            all_homes (bool, optional): TODO determine the affect of this param. Defaults to True.
            sale_options (Set[Literal[ &quot;ForSaleByAgent&quot;, &quot;ForSaleByOwner&quot;, &quot;NewConstruction&quot;, &quot;ComingSoon&quot;, &quot;Auction&quot;, &quot;ForSaleForeclosure&quot;, &quot;RecentlySold&quot; ]], optional): Sale options. Defaults to {"ForSaleByAgent", "ForSaleByOwner", "NewConstruction", "ComingSoon", "Auction", "ForSaleForeclosure"}.
            sold_in_last_x_days (Union[None, int], optional): For looking for already sold houses that have been sold in the last X. Note that this requires 'RecentlySold' in the sale options. This also has no affect if for_sale is False. Defaults to None.
            price_max (Union[None, int], optional): The max price. Defaults to None.
            price_min (Union[None, int], optional): The min price. Defaults to None.
            monthly_payment_max (Union[None, int], optional): The monthly payment max. Defaults to None.
            monthly_payment_min (Union[None, int], optional): The monthly payment min. Defaults to None.
            beds (Union[None, int], optional): The minimum number of beds. Defaults to None.
            baths (Union[None, int], optional): The minimum number of baths. Defaults to None.
            home_type (Set[Literal[ &quot;SingleFamily&quot;, &quot;Townhouse&quot;, &quot;MultiFamily&quot;, &quot;Condo&quot;, &quot;LotLand&quot;, &quot;Apartment&quot;, &quot;Manufactured&quot;, &quot;ApartmentOrCondo&quot;, ]], optional): _description_. Defaults to {"SingleFamily", "Townhouse", "MultiFamily", "Condo", "LotLand", "Apartment", "Manufactured", "ApartmentOrCondo"}.
            hoa (Union[None, int], optional): Types of homes to include. Defaults to None.
            parking_spots (Union[None, int], optional): Minimum number of parking spots. Defaults to None.
            features (Set[Literal[ &quot;Garage&quot;, &quot;BasementFinished&quot;, &quot;BasementUnfinished&quot;, &quot;SingleStory&quot;, &quot;AgeRestricted55Plus&quot;, &quot;AirConditioning&quot;, &quot;Pool&quot;, &quot;Waterfront&quot;, &quot;CityView&quot;, &quot;ParkView&quot;, &quot;MountainView&quot;, &quot;WaterView&quot;, ]], optional): Required features for houses to have in results. Defaults to set().
            house_sqft_max (Union[None, int], optional): House sqft max. Defaults to None.
            house_sqft_min (Union[None, int], optional): House sqft min. Defaults to None.
            lot_sqft_max (Union[None, int], optional): Lot sqft max. Defaults to None.
            lot_sqft_min (Union[None, int], optional): Lot sqft min. Defaults to None.
            year_built_start (Union[None, int], optional): Year built min. Defaults to None.
            year_built_end (Union[None, int], optional): Year built max. Defaults to None.
            keywords (Set[str], optional): Search keywords. Defaults to set().

        Returns:
            'Query': Returns self
        """

        # TODO refactor export logic to helpers and constants

        filter_state = {"isAllHomes": {"value": all_homes}}

        if for_sale:

            possible_sale_options = {"ForSaleByAgent", "ForSaleByOwner", "NewConstruction",
                                     "ComingSoon", "Auction", "ForSaleForeclosure"}

            diff = possible_sale_options - set(sale_options)
            for opt in diff:
                filter_state[f"is{opt}"] = {"value": False}

            if "RecentlySold" in sale_options:
                filter_state["isRecentlySold"] = {"value": True}

                if sold_in_last_x_days:
                    filter_state["doz"] = {"value": str(sold_in_last_x_days)}

        else:
            filter_state["isForRent"] = {"value": True}
            filter_state["isForSaleByAgent"] = {"value": False}
            filter_state["isForSaleByOwner"] = {"value": False}
            filter_state["isNewConstruction"] = {"value": False}
            filter_state["isComingSoon"] = {"value": False}
            filter_state["isAuction"] = {"value": False}
            filter_state["isForSaleForeclosure"] = {"value": False}

        if price_max or price_min:
            price = {}
            if price_max:
                price["max"] = price_max
            if price_min:
                price["min"] = price_min

            filter_state["price"] = price

        if monthly_payment_max or monthly_payment_min:
            monthly = {}
            if monthly_payment_max:
                monthly["max"] = monthly_payment_max
            if monthly_payment_min:
                monthly["min"] = monthly_payment_min

            filter_state["monthlyPayment"] = monthly

        if beds:
            filter_state["beds"] = {"min": beds}

        if baths:
            filter_state["baths"] = {"min": baths}

        if home_type:
            possible_home_types = {"SingleFamily", "Townhouse", "MultiFamily",
                                   "Condo", "LotLand", "Apartment", "Manufactured", "ApartmentOrCondo"}

            diff = possible_home_types - set(home_type)
            for opt in diff:
                filter_state[f"is{opt}"] = {"value": False}

        if hoa:
            filter_state["hoa"] = {"max": hoa}

        if parking_spots:
            filter_state["parkingSpots"] = {"min": parking_spots}

        if features:
            feature_dict = {
                "Garage": "hasGarage",
                "BasementFinished": "isBasementFinished",
                "BasementUnfinished": "isBasementUnfinished",
                "SingleStory": "singleStory",
                "AgeRestricted55Plus": "ageRestricted55Plus",
                "AirConditioning": "hasAirConditioning",
                "Pool": "hasPool",
                "Waterfront": "isWaterfront",
                "CityView": "isCityView",
                "ParkView": "isParkView",
                "MountainView": "isMountainView",
                "WaterView": "isWaterView",
            }

            for f in features:
                if f_key := feature_dict.get(f, None):
                    filter_state[f_key] = {"value": True}

        if house_sqft_max or house_sqft_min:
            house_sqft = {}
            if house_sqft_max:
                house_sqft["max"] = house_sqft_max
            if house_sqft_min:
                house_sqft["min"] = house_sqft_min

            filter_state["sqft"] = house_sqft

        if lot_sqft_max or lot_sqft_min:
            lot_sqft = {}
            if lot_sqft_max:
                lot_sqft["max"] = lot_sqft_max
            if lot_sqft_min:
                lot_sqft["min"] = lot_sqft_min

            filter_state["lotSize"] = lot_sqft

        if year_built_start or year_built_end:
            year_range = {}
            if year_built_start:
                year_range["min"] = year_built_start
            if year_built_end:
                year_range["max"] = year_built_end

            filter_state["built"] = year_range

        if keywords:
            filter_state["keywords"] = {
                "value": f"{' ,'.join(keywords)}"
            }

        self.sub_parms["filterState"] = filter_state

        return self

    def set_list_visable(self, list_visible: bool) -> 'Query':
        # TODO determine what impact this has
        self.sub_parms["isListVisible"] = list_visible
        return self

    def set_map_zoom(self, zoom: int) -> 'Query':
        # TODO determine what impact this has
        self.sub_parms["mapZoom"] = zoom
        return self

    def set_wants(self, wants: dict) -> 'Query':
        # TODO make this more user friendly and determine the effect of wants
        self.wants = wants
        return self

    def set_custom_region(self, region: str, url: str = defaults.URL, headers: Dict = defaults.HEADER) -> 'Query':
        """Sets a custom region to limit query results to. This is accomplished by sending a POST request to Zillow with the region string.

        Args:
            region (str): A string of longitudes and latitudes in the format of 'long1,lat1|long2,lat3|...|long1,lat1'. Note that the last pair of longitude and latitude values should be the same the the first.
            url (str, optional): The zillow URL for the POST request. The default can be found within defaults.py within the Zillow module.
            headers (Dict, optional): The headers sent with the POST request. The default can be found within defaults.py within the Zillow module.

        Returns:
            Query: Returns self
        """
        payload = f"-----011000010111000001101001\r\nContent-Disposition: form-data; name=\"clipPolygon\"\r\n\r\n{region}\r\n-----011000010111000001101001--\r\n"

        response = requests.request("POST", url, data=payload, headers=headers)
        region_id = response.json()["customRegionId"]

        self.sub_parms["customRegionId"] = region_id
        return self

    def clear_custom_region(self) -> 'Query':
        """Clears the custom region ID from the query.

        Returns:
            Query: Returns self
        """
        self.sub_parms.pop("customRegionId")
        return self

    def clear_filter(self) -> 'Query':
        """Clears the filter state.

        Returns:
            Query: Returns self
        """
        self.sub_parms["filterState"] = {}
        return self

    def get_params_string(self) -> str:
        """Generates the parameters JSON string according to the values set within the query

        Returns:
            str: The parameters JSON string
        """
        return {
            "searchQueryState": json.dumps(self.sub_parms),
            "wants": json.dumps(self.wants)
        }

    def get_response(self, returns: Literal["request", "full", "results"] = "request", url: str = defaults.URL, headers: Dict = defaults.HEADER) -> Any:
        """Returns the requests response object for the query with the configured params

        Args:
            returns ('request' | 'full' | 'results'): Sets what this should return, request -> request return object, 'full' -> full response json, 'results' -> search results list. Defaults to 'request'.
            url (str, optional): The Zillow URL for the API request. The default can be found within defaults.py within the Zillow module
            headers (Dict, optional): The headers parameters as a dictionary. The Defaults can be found within defaults.py within the Zillow module.

        Returns:
            Any: Returns either a requests response object or dict or list. This is dependent on the returns parameter.
        """

        r = requests.request(
            "GET",
            url,
            headers=headers,
            params=self.get_params_string()
        )

        if returns == "request":
            return r
        if returns == "full":
            return r.json()
        if returns == "results":
            return r.json().get("cat1").get('searchResults').get('listResults')



# This handles parsing of rental apartments data
import requests
from typing import Dict, Any, List
from numbers import Number
from .details_page import NextJS_Detail_Page
from .. import defaults


class Rental_Apartment(NextJS_Detail_Page):

    def __init__(self, url: str) -> None:
        """This initializes the Rental_Apartment object, which call a GET request on the Zillow detail URL. 

        Attributes:
            url (str): The detail URL that this class has parsed
            soup (BeautifulSoup): The html content of this page parsed to a soup object

            ndata (Dict): The __NEXT_DATA__ data dictionary scraped from html soup
            idata (Dict): The initial data dictionary component of ndata
            redux_state (Dict): the redux state dictionary component of ndata

            zpid (str): The Zillow property ID
            lot_id (str): The Zillow lot ID

            data (Dict): The API data dictionary. Will be result of a fresh API call unless no data is returned, in which case will fall back to idata (initial data)

            building (Dict): The building data dictionary, found within data
            building_attributes (Dict): The building attributes data dictioanry, found within building

            building_name (str): The building name

            description (str): The apartment building description
            low_income (bool): If apartment building is low income
            senior_housing (bool): If building is senior housing
            student_housing (bool): If building is student housing

            office_hours (List[str]): List of the building's office hours
            office_number (str): The phone number of the building's office

            key_features (Dict): Key features of the building
            unit_features (List[str]): The universal features of the apartment units

            city (str): The city
            county (str): The county
            state (str): The state
            zip (str): The zip code
            street_address (str): The street address of the building

            application_fee (Number | None): The application fee
            administrative_fee (Number | None): The administrative fee 
            deposit_fee_min (Number | None): The min deposit fee
            deposit_fee_max (Number | None): The max deposit fee
            lease_terms (List[str]): The lease teams. Note that is is more a keyword/descriptor as opposed to a legal lease contract terms. So you might expected something like ['Flexable'] if the terms are flexable/negotiable
            utilities_included (List[str]): The included utilities

            parking_policies (List[dict]): List of parking policies
            parking_types (List[str]): List of parking types available, for example "off_road"
            pet_policies (List[dict]): The list of pet policies

            shared_laundry (bool): Is there shared laundry, True/False
            air_conditioning (str): Air conditioning description
            appliances (List[str]): List of appliances
            outdoor_common_areas (List[str]): List of outdoor common areas
            barbecue (bool): If there a/are barbecue(s) in the common areas, True/False
            heating_source (str): The heating source
            elevator (Any): The elevator specification of the building
            community_rooms (List[str]): List of community rooms available
            sports_courts (List[str]): List of sports courts available
            bicycle_storage (Any | None): Bicycle storage specification
            guest_suite (Any | None): Guest Suite specification
            storage (Any | None): Storage specification
            pet_park (Any | None): Pet park specification
            maintenance_24_7 (Any | None): Specification of 24/7 maintenance availability
            dry_cleaning_drop_off (Any | None): Specification of dry cleaning drop off availability
            online_rent_payment (Any | None): Specification of online rent payment availability
            online_maintenance_portal (Any | None): Specification of online maintenance portal availability
            onsite_management (bool): If management is onsite, True/False
            package_service (Any | None): Specification of package service availability
            valet_trash (Any | None): Specification of valet trash availability
            spanish_speaking_staff (Any | None): Specification of spanish speaking staff availability
            security_types (List[str]): The types of security that building has
            view_types (List[str]): List of views (scenic) that the building has
            hot_tub (Any | None): Specification of hot tub availability
            sauna (Any | None): Specification of sauna availability
            swimming_pool (bool): If has swimming pool, True/False
            assisted_living (bool): If building has assisted living, True/False
            disabled_access (Any | None): The specification of disabled access availablity
            floor_covering (List[str]): The list of floor coverings that the building has
            communication_types (List[str]): The list of communication types the building has/supports. For instance "high speed internet"
            ceiling_fan (Any | None): Specification of ceiling fan availability
            fire_place (Any | None): Specification of fire place availability
            patio_balcony (bool): If building has patio balcony, True/False
            furnished (Any | None): Specification of if/how is furnished
            custom_ammenites (str): Description of custom ammenites

            floorplans (List[dict]): List of available floor plans

            management (dict): Management details

            schools (List[dict]): The list of nearby schools
            nearby_cities (List[dict]): The list of nearby cities
            nearby_neighborhoods (List[dict]): The list of nearby neighborhoods
            nearby_zip (List[dict]): The list of nearby zip codes
            nearby_rental_buildings (List[dict]): The list of nearby rental homes and apartments
            nearby_amenities (List[dict]): The list of nearby amenities 

            review_info (dict): The data dictionary of reviews
        """

        self.url = url

        # get soup
        self.soup = self.make_soup(self.get_page(self.url))

        # get ndata and initial
        self.ndata = self.get_next_data(self.soup)
        self.idata, self.redux_state = self.get_initial_data_and_redux_state(
            self.ndata)

        # get ids
        self.zpid: str = self.idata['building']['zpid']
        self.lot_id: str = self.idata['building']['lotId']

        # if fresh data available, get fresh
        self.data = self.get_fresh_graphQL_data(self.lot_id)
        if not self.data:  # if empty dictionary fall back to initial data
            self.data = self.idata

        self.building: Dict[str, Any] = self.data['building']
        self.building_attributes: Dict[str,
                                       Any] = self.building['buildingAttributes']
        self.building_name: str = self.building['buildingName']

        self.description: str = self.building['description']
        self.low_income: bool = self.building['isLowIncome']
        self.senior_housing: bool = self.building['isSeniorHousing']
        self.student_housing: bool = self.building['isStudentHousing']

        self.office_hours: List[str] = self.building['amenityDetails']['hours']
        self.office_number: str = self.building['buildingPhoneNumber']

        self.key_features = self.get_key_features()
        self.unit_features: List[str] = self.building['amenityDetails']['unitFeatures']

        self.city: str = self.building['city']
        self.county: str = self.building['county']
        self.state: str = self.building['state']
        self.zip: str = self.building['zipcode']
        self.street_address: str = self.building['fullAddress']

        self.application_fee: Number | None = self.building_attributes['applicationFee']
        self.administrative_fee: Number | None = self.building_attributes['administrativeFee']
        self.deposit_fee_min: Number | None = self.building_attributes['depositFeeMin']
        self.deposit_fee_max: Number | None = self.building_attributes['depositFeeMax']
        self.lease_terms: List[str] = self.building_attributes.get(
            'leaseTerms', [])
        self.utilities_included: List[str] = self.building_attributes.get(
            'utilitiesIncluded', [])

        self.parking_policies: List[Dict[str, Any]
                                    ] = self.building_attributes['detailedParkingPolicies']
        self.parking_types: List[str] = self.building_attributes['parkingTypes']

        self.pet_policies: List[Dict[str, Any]
                                ] = self.building_attributes['detailedPetPolicy']

        self.shared_laundry: bool = self.building_attributes['hasSharedLaundry']
        self.air_conditioning: str = self.building_attributes['airConditioning']
        self.appliances: List[str] = self.building_attributes['appliances']
        self.outdoor_common_areas: List[str] = self.building_attributes['outdoorCommonAreas']
        self.barbecue: bool = self.building_attributes['hasBarbecue']
        self.heating_source: str = self.building_attributes['heatingSource']
        self.elevator: Any | None = self.building_attributes['hasElevator']
        self.community_rooms: List[str] = self.building_attributes['communityRooms']
        self.sports_courts: List[str] = self.building_attributes['sportsCourts']
        self.bicycle_storage: Any | None = self.building_attributes['hasBicycleStorage']
        self.guest_suite: Any | None = self.building_attributes['hasGuestSuite']
        self.storage: Any | None = self.building_attributes['hasStorage']
        self.pet_park: Any | None = self.building_attributes['hasPetPark']
        self.maintenance_24_7: Any | None = self.building_attributes[
            'hasTwentyFourHourMaintenance']
        self.dry_cleaning_drop_off: Any | None = self.building_attributes[
            'hasDryCleaningDropOff']
        self.online_rent_payment: Any | None = self.building_attributes['hasOnlineRentPayment']
        self.online_maintenance_portal: Any | None = self.building_attributes[
            'hasOnlineMaintenancePortal']
        self.onsite_management: bool = self.building_attributes['hasOnsiteManagement']
        self.package_service: Any | None = self.building_attributes['hasPackageService']
        self.valet_trash: Any | None = self.building_attributes['hasValetTrash']
        self.spanish_speaking_staff: Any | None = self.building_attributes[
            'hasSpanishSpeakingStaff']
        self.security_types: List[str] = self.building_attributes['securityTypes']
        self.view_types: List[str] = self.building_attributes['viewType']
        self.hot_tub: Any | None = self.building_attributes['hasHotTub']
        self.sauna: Any | None = self.building_attributes['hasSauna']
        self.swimming_pool: bool = self.building_attributes['hasSwimmingPool']
        self.assisted_living: bool = self.building_attributes['hasAssistedLiving']
        self.disabled_access: Any | None = self.building_attributes['hasDisabledAccess']
        self.floor_covering: List[str] = self.building_attributes['floorCoverings']
        self.communication_types: List[str] = self.building_attributes['communicationTypes']
        self.ceiling_fan: Any | None = self.building_attributes['hasCeilingFan']
        self.fire_place: Any | None = self.building_attributes['hasFireplace']
        self.patio_balcony: bool = self.building_attributes['hasPatioBalcony']
        self.furnished: Any | None = self.building_attributes['isFurnished']
        self.custom_ammenites: str = self.building_attributes['customAmenities']

        self.floorplans: List[Dict[str, Any]] = self.building['floorPlans']

        self.management: Dict[str, Any] = self.get_management_company(
            self.zpid).get('rentalListingOwnerContact', {})

        self.schools: List[Dict[str, Any]
                           ] = self.building.get('assignedSchools', [])
        self.nearby_cities: List[Dict[str, Any]
                                 ] = self.building.get('nearbyCities', [])
        self.nearby_neighborhoods: List[Dict[str, Any]
                                        ] = self.building.get('nearbyNeighborhoods', [])
        self.nearby_zip: List[Dict[str, Any]
                              ] = self.building.get('nearbyZipcodes', [])
        self.nearby_rental_buildings: List[Dict[str, Any]
                                           ] = self.building.get('nearbyBuildingLinks', [])
        self.nearby_amenities: List[Dict[str, Any]
                                    ] = self.building.get('nearbyAmenities', [])

        self.review_info: Dict[str, Any] = self.building.get('reviewsInfo', {})

    @staticmethod
    def get_fresh_graphQL_data(lot_id: str) -> Dict[str, str]:
        """Sends a GRAPHQL query to get the fresh data, as opposed to the initial data which may be stale or incomplete (for instance, lease terms are missing in initial data)

        Args:
            lot_id (str): The lot id

        Returns:
            Dict[str, str]: The data dictionary
        """

        url = "https://www.zillow.com/graphql"

        payload = {
            "operationName": "BuildingQuery",
            "variables": {
                "cache": False,
                "lotId": lot_id,
                "update": False},
            "queryId": "efbb40baf8ba7747347be4d8b170edc9"
        }

        return requests.request(
            "POST", url, json=payload, headers=defaults.GRAPHQL_HEADER
        ).json().get('data', {})

    def get_key_features(self) -> Dict[str, str]:
        """Gets the key features dictionary from the detail page.

        Returns:
            Dict[str, str]: Key features
        """
        return {
            t.title.text: t.span.text for t in self.soup.find(
                "div", id="bdp-building-facts-and-features"
            ).ul.find_all('li')
        }

    @staticmethod
    def get_management_company(zpid: str) -> Dict[str, Any]:
        """Gets the management details via a GRAPHQL query

        Args:
            zpid (str): The zillow ID

        Returns:
            Dict[str, Any]: The management details JSON
        """

        url = "https://www.zillow.com/graphql"

        payload = {
            "operationName": "ListingContactDetailsQuery",
            "variables": {"zpid": zpid},
            "query": r"query ListingContactDetailsQuery($zpid: ID!) {  viewer {    roles {      isLandlordLiaisonMember      isLlpRenter      __typename    }    __typename  }  property(zpid: $zpid) {    zpid    brokerId    isHousingConnector    isIncomeRestricted    rentalListingOwnerReputation {      responseRate      responseTimeMs      contactCount      applicationCount      isLandlordIdVerified      __typename    }    isFeatured    isListedByOwner    rentalListingOwnerContact {      displayName      businessName      phoneNumber      agentBadgeType      photoUrl      reviewsReceivedCount      reviewsUrl      ratingAverage      isBrokerLocalCompliance      __typename    }    postingProductType    postingContact {      brokerName      brokerageName      name      __typename    }    postingUrl    rentalMarketingTreatments    building {      bdpUrl      buildingName      housingConnector {        hcLink {          text          __typename        }        __typename      }      ppcLink {        text        __typename      }      __typename    }    __typename  }}"
        }

        return requests.request(
            "POST", url, json=payload, headers=defaults.GRAPHQL_HEADER,)\
            .json().get('data', {}).get('property', {})



# this file is responsible for parsing the detailUrl page found in the query results.
from typing import Literal, List, Dict, Any
import numpy as np
from time import sleep
from numbers import Number
from . import Sale, Rental_Home, Rental_Apartment


def scrape_listing(detail_url: str, status_type: Literal["FOR_RENT", "FOR_SALE"]) -> Sale | Rental_Apartment | Rental_Home:
    """Scrapes a listing based on the Zillow Detail URL

    Args:
        detail_url (str): The URL with the listing detaill
        status_type (Literal[&quot;FOR_RENT&quot;, &quot;FOR_SALE&quot;]): The status type, which should be either FOR_RENT or FOR_SALE

    Raises:
        ValueError: Detail URL for FOR_RENT is missing expected keywords
        ValueError: Unhandled status_type/unexpected status_type

    Returns:
        Sale | Rental_Apartment | Rental_Home: _description_
    """

    if status_type == "FOR_SALE":
        return Sale(detail_url)

    if status_type == "FOR_RENT":
        if 'homedetails' in detail_url:
            return Rental_Home(detail_url)
        elif '/b/' in detail_url:
            return Rental_Apartment(f"https://www.zillow.com{detail_url}")
        else:
            raise ValueError(
                "FOR_RENTAL listing detail URL is missing expected keywords")

    raise ValueError(
        f"status_type should be either FOR_RENT or FOR_SALE, was {status_type}")


def scrape_listings(query_results: List[Dict[str, Any]], delay: Number = 0, jitter: Number = 1, verbose=False) -> List[Sale] | List[Rental_Home] | List[Rental_Apartment]:
    """Scrapes a list of Zillow Detail URLs

    Args:
        query_results (List[Dict[str, Any]]): Results of the Query, see Query class
        delay (Number, optional): The static delay amount. This is to help prevent Zillow from blocking scraping due to high number of requests in rapid succession. Defaults to 0.
        jitter (Number, optional): The jitter factor for the delay time. Defaults to 1.
        verbose (bool, optional): If function should be verbose, printing out the progress of parsing the listings. Defaults to False.

    Returns:
        List[Sale] | List[Rental_Home] | List[Rental_Apartment]: List of scraped listing, which typically will be of one type, Sale, Rental_home, or Rental_Apartment, that being said if multiple types are mixed in query_results, the resulting list will be of multiple types.
    """

    delay_array = delay * np.random.rand(len(query_results)) * jitter
    # last delay should be zero since no more details to scrape
    delay_array[-1] = 0

    results = []
    for dtime, (i, result) in zip(delay_array, enumerate(query_results, 1)):

        results.append(
            scrape_listing(result["detailUrl"], result['statusType'])
        )

        if verbose:
            print(f"Parsed {i} of {len(query_results)}, delaying for {dtime}s")

        sleep(dtime)

    return results


class Lazy_Listings(list):
    """Lazy Listings is a list child class that takes a list of Query results and will return the scraped listing details based on the index or slice.
    """

    def __getitem__(self, n):
        result = super().__getitem__(n)

        if isinstance(result, list):
            return [scrape_listing(r["detailUrl"], r['statusType']) for r in result]
        else:
            return scrape_listing(result["detailUrl"], result['statusType'])

    def __iter__(self):
        for result in super().__iter__():
            yield scrape_listing(result["detailUrl"], result['statusType'])


def lazy_scrape_listings(query_results: List[Dict[str, Any]]) -> Lazy_Listings:
    """Scapes Zillow listing details with Lazy Evaluation. The Lazy_Listing class it returns can be used to scrape the details of the query results list by either iteration or indexing/slicing.

    Args:
        query_results (List[Dict[str, Any]]): Results of the Query, see Query class

    Returns:
        Lazy_Listings: Lazy_Listing list of query_results
    """
    return Lazy_Listings(query_results)



import requests
import json
import requests
from typing import Dict, Any, List, Literal, Set
from datetime import datetime
from . import defaults


class Sale_Query:

    def __init__(self) -> None:

        self.set_post_params()
        self.set_json_payload()
        self.set_graphql_query()

    def set_post_params(self, post_parms: Dict[str, Any] = {"client_id": "rdc-x", "schema": "vesta"}) -> 'Sale_Query':
        """Sets the post request parameters, the dictionary that will be passed to the 'param' parameter in the post request call. The exact effect of these is not entirely clear and therefore it is recommended to left as default.

        Args:
            post_parms (Dict[str, Any], optional): The dictionary of post request paramters. Defaults to {"client_id": "rdc-x", "schema": "vesta"}.

        Returns:
            Query: Returns self
        """

        if not post_parms:
            self.post_params = {"client_id": "rdc-x", "schema": "vesta"}
        else:
            self.post_params = post_parms

        return self

    def set_json_payload(self, payload: Dict[str, Any] = defaults.LISTING_SEARCH_PAYLOAD) -> 'Sale_Query':
        """Sets the request payload, this is the body of the request, sent as a POST. Note that this is rather complicated as the Realotr PUBLIC API is not documented therefore it is recommended to use the default. Also when setting the payload, there are two key keys that are missing real values, 'query' (GraphQL query - see set_graphql_query) and 'variables' --> 'query' (fitler criteria - see set_filter_query and set_filter_query_preset)

        Args:
            payload (Dict[str, Any], optional): The POST request JSON payload. Defaults to defaults.LISTING_SEARCH_PAYLOAD.

        Returns:
            Query: Returns self
        """

        self.payload = payload
        return self

    def set_payload_variables(self, **kwargs) -> 'Sale_Query':
        """Sets variable key value pairs inside the POST request payload. This is not recommended to do manually, for advanced use only.

        Returns:
            Query: Returns self
        """

        for key, item in kwargs.items():
            self.payload['variables'][key] = item

        return self

    def remove_payload_variables(self, *args) -> 'Sale_Query':
        """Removes variable keys inside the POST request payload. This is not recommended to do manually (as in using this function), for advanced use only.

        Returns:
            Query: Returns self
        """

        for key in args:
            if key in self.payload['variables']:
                self.payload['variables'].pop(key)

        return self

    def set_var_client_data(self, client_data: Dict[str, Any] = {"device_data": {"device_type": "web"}, "user_data": {}}) -> 'Sale_Query':
        """Sets the 'client_data' var in payload variables. The exact effect of this is not clear, but if this variable is excluded, the response will fail.

        Args:
            client_data (Dict[str, Any]): The client_data dictionary. Defaults to {"device_data": {"device_type": "web"}, "user_data": {}}

        Returns:
            Query: Returns self
        """

        if client_data is None:
            return self.remove_payload_variables('client_data')

        self.payload['variables']['client_data'] = client_data
        return self

    def set_var_limit(self, limit: int = 42) -> 'Sale_Query':
        """Sets the limit variable in payload variables. This is the max number of results that will be returned. Note that if a query has less results than the limit, then the number of results will be less than the limit.

        Args:
            limit (int, optional): The limit. Defaults to 42, which is the realtor.com default.

        Returns:
            Query: Return self
        """

        if limit is None:
            return self.remove_payload_variables('limit')

        self.payload['variables']['limit'] = limit
        return self

    def set_var_offset(self, offset: int = 0) -> 'Sale_Query':
        """Sets the result offset in the payload variables. This is essentially how pagination works in realtor.com, for instance if the limit is 42, page 2 starts on offset=42 (counting from 0).

        Args:
            offset (int, optional): The result offset. Defaults to 0.

        Returns:
            Query: Returns self
        """

        if offset is None:
            return self.remove_payload_variables('offset')

        self.payload['variables']['offset'] = offset
        return self

    def set_var_sort_type(self, sort_type: str | Dict[str, Any]) -> 'Sale_Query':
        """Sets the sort_type in the payload variables. Note that this is mutually exclusive with sort_type. This is an advanced function, for regular usage, use set_var_sort_preset.

        Args:
            sort_type (str | Dict[str, Any]): The sort_type paramter that specifies how the results will be sorted

        Returns:
            Query: Returns self
        """

        if sort_type is None:
            return self.remove_payload_variables('sort_type')

        self.payload['variables']['sort_type'] = sort_type
        return self

    def set_var_sort(self, sort: Dict[str, Any]) -> 'Sale_Query':
        """Sets the sort in payload variables. Note that this is mutually exclusive with sort_type. This function is for advanced usage only, set_var_sort_preset is recommended for regular usage.

        Args:
            sort (Dict[str, Any]): The sort parameter dictionary

        Returns:
            Query: Returns self
        """
        if sort is None:
            return self.remove_payload_variables('sort')

        self.payload['variables']['sort'] = sort
        return self

    def set_var_sort_preset(self, by: Literal["relevant", "price", "listing age", "open house date", "last reduced", "interior sqft", "lot_size"] = "relevant", ascending=False) -> 'Sale_Query':
        """Sets the sorting and therefore the order of the results.

        Args:
            by (Literal[&quot;relevant&quot;, &quot;price&quot;, &quot;listing age&quot;, &quot;open house date&quot;, &quot;last reduced&quot;, &quot;interior sqft&quot;, &quot;lot_size&quot;], optional): The criteria by which to sort. Defaults to "relevant".
            ascending (bool, optional): If results should be sorted in ascending order or (False) descening. Note that this doesn't have an effect on 'relevant'. Defaults to False.

        Returns:
            Query: Returns self
        """

        sort_direction = 'asc' if ascending else 'desc'

        if by == "relevant":
            self.set_var_sort(None)
            return self.set_var_sort_type('relevant')
        elif by == "price":
            self.set_var_sort_type(None)
            return self.set_var_sort(
                {
                    "field": "list_price",
                    "direction": sort_direction,
                }
            )
        elif by == "listing age":
            self.set_var_sort_type(None)
            return self.set_var_sort(
                {
                    "field": "list_date",
                    "direction": sort_direction,
                }
            )
        elif by == "open house date":
            self.set_var_sort_type(None)
            return self.set_var_sort(
                {
                    "field": "open_house_date",
                    "direction": sort_direction,
                }
            )
        elif by == "last reduced":
            self.set_var_sort_type(None)
            return self.set_var_sort(
                {
                    "field": "price_reduced_date",
                    "direction": sort_direction,
                }
            )
        elif by == "interior sqft":
            self.set_var_sort_type(None)
            return self.set_var_sort(
                {
                    "field": "sqft",
                    "direction": sort_direction,
                }
            )
        elif by == "lot_size":
            self.set_var_sort_type(None)
            return self.set_var_sort(
                {
                    "field": "lot_sqft",
                    "direction": sort_direction,
                }
            )
        else:
            raise ValueError("Invalid 'by' value")

    def set_var_zoho_query(self, zoho_query: Dict[str, Any]) -> 'Sale_Query':
        """Sets the zohoQuery in payload variables. The effect of this variable is unclear and is not required. For advanced usage only. This function was only included because this variable is used by realtor.com web, although it is excluded from the defaults in the 'Query' class.

        Args:
            zoho_query (Dict[str, Any]): The zohoQuery

        Returns:
            Query: Returns self
        """
        if zoho_query is None:
            return self.remove_payload_variables('zohoQuery')

        self.payload['variables']['zohoQuery'] = zoho_query
        return self

    def set_var_geo_supported_slug(self, geo_supported_slug: Dict[str, Any]) -> 'Sale_Query':
        """Sets the geoSupportedSlug in payload variables. The effect of this variable is unclear and is not required. For advanced usage only. This function was only included because this variable is used by realtor.com web, although it is excluded from the defaults in the 'Query' class.

        Args:
            geo_supported_slug (Dict[str, Any]): The geoSupportedSlug

        Returns:
            Query: Returns self
        """
        if geo_supported_slug is None:
            return self.remove_payload_variables('geoSupportedSlug')

        self.payload['variables']['geoSupportedSlug'] = geo_supported_slug
        return self

    def set_var_by_prop_type(self, by_prop_type: Dict[str, Any]) -> 'Sale_Query':
        """Sets the by_prop_type in payload variables. The effect of this variable is unclear and is not required. For advanced usage only. This function was only included because this variable is used by realtor.com web, although it is excluded from the defaults in the 'Query' class.

        Args:
            by_prop_type (Dict[str, Any]): The by_prop_type

        Returns:
            Query: Returns self
        """
        if by_prop_type is None:
            return self.remove_payload_variables('by_prop_type')

        self.payload['variables']['by_prop_type'] = by_prop_type
        return self

    def set_graphql_query(self, query: str = defaults.GRAPHQL_LISTING_SEARCH_QUERY) -> 'Sale_Query':
        """Sets the GraphQL query. Note that this is rather complicated as the Realtor PUBLIC API is not documented therefore it is recommended to use the default.

        Args:
            query (str, optional): The GraphQL query for the listings search. Defaults to defaults.GRAPHQL_LISTING_SEARCH_QUERY.

        Returns:
            Query: Returns self
        """

        self.payload['query'] = query
        return self

    def set_filter_query(self, query: Dict[str, Any]) -> 'Sale_Query':
        """Sets the filter query within 'payload' -> 'variables' -> 'query'. This is used to filter which results should be returned by the search. Note that this is a complex function to use as the PUBLIC Realor.com isn't documented therefore using set_filter_query_preset is recommended as it builds a filter query based on simple parameters.

        Args:
            query (Dict[str, Any]): The filter for the listings query

        Returns:
            Query: Returns self
        """

        self.payload['variables']['query'] = query
        return self

    def set_filter_query_preset(  # search loc, cities, and state code mutually exclusive
        self,
        search_location: str = None,
        location_radius: int = None,  # expand by x miles, only if search_location
        # cluster of one or more cities to include
        cities: List[Dict[str, str]] = None,
        state_code: str = None,
        primary: bool = True,
        new_construction: bool = None,
        foreclosure: bool = None,
        senior_community: bool = None,  # 55+ age
        contingent: bool = None,
        has_tour: bool = None,  # virtual toor
        single_story: bool = None,
        multi_story: bool = None,
        open_house_date_min: datetime = None,
        open_house_date_max: datetime = None,
        list_date_min: datetime = None,
        list_date_max: datetime = None,
        price_min: int = None,
        price_max: int = None,
        price_reduce_date_min: datetime = None,
        price_reduce_date_max: datetime = None,
        hoa_fee_min: int = None,
        hoa_fee_max: int = None,
        # if should show properties with incomplete HOA data
        include_incomplete_hoa: bool = True,
        bedrooms_min: int = None,
        bedrooms_max: int = None,
        bathrooms_min: int = None,
        bathrooms_max: int = None,
        interior_sqft_min: int = None,
        interior_sqft_max: int = None,
        lot_sqft_min: int = None,
        lot_sqft_max: int = None,
        year_built_min: int = None,
        year_built_max: int = None,
        min_garages: int = 0,
        has_carport: bool = None,
        has_rv_or_boat_parking: bool = None,
        has_central_air: bool = None,
        has_forced_air: bool = None,
        has_central_heat: bool = None,
        has_basement: bool = None,
        has_hardwood_floors: bool = None,
        has_fireplace: bool = None,
        has_disability_features: bool = None,
        has_dining_room: bool = None,
        has_family_room: bool = None,
        has_den_or_office: bool = None,
        has_swimming_pool: bool = None,
        has_spa_or_hot_tub: bool = None,
        has_horse_facilities: bool = None,
        has_hill_or_mountain_view: bool = None,
        has_ocean_view: bool = None,
        has_lake_view: bool = None,
        has_river_view: bool = None,
        has_golf_course_lot_or_frontage: bool = None,
        has_corner_lot: bool = None,
        is_cul_de_sac: bool = None,
        has_waterfront: bool = None,
        has_community_swimming_pool: bool = None,
        has_community_spa_or_hot_tub: bool = None,
        has_community_golf: bool = None,
        has_community_security_features: bool = None,
        has_community_clubhouse: bool = None,
        has_tennis_court: bool = None,
        has_community_boat_facilities: bool = None,
        status: Set[Literal[
            "for_sale",
            "ready_to_build",
            "sold"

        ]] = {"for_sale"},
        sold_date_min: datetime = None,
        sold_date_max: datetime = None,
        prop_type: Set[Literal[
            "farm",
            "mobile",
            "multi_family",
            "townhomes",
            "duplex_triplex",
            "single_family",
            "condos",
            "condo_townhome_rowhome_coop",
            "condo_townhome",
            "land"
        ]] = None,  # all -> not specified in query
        keywords: List[str] = None,
        tags: List[str] = [],
        exclude_tags: List[str] = []
    ) -> 'Sale_Query':
        """Sets the filter based on the parameters provided.

        Args:
            search_location (str, optional): The location to search, for example "Charolette, NC". Mutally exclusive with with 'cities'. Defaults to None. NOTE: Either search_location or cities must be specified for query to succeed
            location_radius (int, optional): The radius in miles around the search_location to expand the search to. This is in addition to the Realtor.com default scope of the search, so 0 gives the results without any expansion of the search area/circle. Defaults to 0
            cities (List[Dict[str, str]], optional): List of cities to include in the search. This is mutally exclusive with 'search_location'. Cities are to be specified as [{"city": "CITY_NAME", "state_code": "STATE_CODE}]. NOTE: Either cities or search_location must be specified for query to succeed
            state_code (str, optional): The state code to filter result to. Defaults to None
            primary (bool): Unclear what this represents. Realtor.com queries use primary=True. Defaults to True
            new_construction (bool, optional): If new constructions should be included or excluded. Defaults to None
            foreclosure (bool, optional): If foreclosures should be included or excluded. Default to None
            senior_community (bool, optional): If senior_community (55+) properties should be included or excluded. Defaults to None
            contingent (bool, optional): If properties that have a contingent status should be included or excluded. Defaults to None
            has_tour (bool, optional): If properties filter properties on if they have a tour or if they do not. Defaults to None
            single_story (bool, optional): If properties should only be single story. Defaults to None
            multi_story (bool, optional): If properties should only be multi story. Defaults to None
            open
            open_house_date_min (datetime, optional): Filter based on start date range of open houses. Defaults to None
            open_house_date_max (datetime, optional): Filter based on end date range of open houses. Defaults to None
            list_date_min (datetime, optional): Filter based on the minumum (at least as old as) listing date. Defaults to None
            listing_date_max (datetime, optional): Filter absed on the maximum (at most as young as) listing date. Defaults to None
            price_min (int, optional): The minumum price. Defaults to None
            price_max (int, optional): The maximum price. Defaults to None
            hoa_fee_min (int, optional): The minumum hoa fee. Defaults to None
            hoa_fee_max (int, optional): The maximum hoa fee. Defaults to None
            include_incomplete_hoa (bool, optional): If houses with incomplete (ergo likely inaccurate) hoa information should be included or filtered out. Defaults to None
            bedrooms_min (int, optional): The minumum number of bedrooms. Defaults to None
            bedrooms_max (int, optional): The maximum number of bedrooms. Defaults to None
            bathrooms_min (int, optional): The minumum number of bathrooms. Defaults to None
            bathrooms_max (int, optional): The maximum number of bathrooms. Defaults to None
            interior_sqft_min (int, optional): Interior sqft min. Defaults to None
            interior_sqft_max (int, optional): Interior sqft max. Defaults to None
            lot_sqft_min (int, optional): Lot sqft min. Defaults to None
            lot_sqft_max (int, optional): Lot sqft max. Defaults to None
            year_built_min (int, optional): Year built min. Defaults to None
            year_built_max (int, optional): Year built max. Defaults to None
            min_garages (int, optional): The minimum garages/number of cars that can fit in garage. Defaults to None
            has_carport (bool, optional): Defaults to None
            has_rv_or_boat_parking (bool, optional): Defaults to None
            has_central_air (bool, optional): Defaults to None
            has_forced_air (bool, optional): Defaults to None
            has_central_heat (bool, optional): Defaults to None
            has_basement (bool, optional): Defaults to None
            has_hardwood_floors (bool, optional): Defaults to None
            has_fireplace (bool, optional): Defaults to None
            has_disability_features (bool, optional): Defaults to None
            has_dining_room (bool, optional): Defaults to None
            has_family_room (bool, optional): Defaults to None
            has_den_or_office (bool, optional): Defaults to None
            has_swimming_pool (bool, optional): Defaults to None
            has_spa_or_hot_tub (bool, optional): Defaults to None
            has_horse_facilities (bool, optional): Defaults to None
            has_hill_or_mountain_view (bool, optional): Defaults to None
            has_ocean_view (bool, optional): Defaults to None
            has_lake_view (bool, optional): Defaults to None
            has_river_view (bool, optional): Defaults to None
            has_golf_course_lot_or_frontage (bool, optional): Defaults to None
            has_corner_lot (bool, optional): Defaults to None
            is_cul_de_sac (bool, optional): Defaults to None
            has_waterfront (bool, optional): Defaults to None
            has_community_swimming_pool (bool, optional): Defaults to None
            has_community_spa_or_hot_tub (bool, optional): Defaults to None
            has_community_golf (bool, optional): Defaults to None
            has_community_security_features (bool, optional): Defaults to None
            has_community_clubhouse (bool, optional): Defaults to None
            has_tennis_court (bool, optional): Defaults to None
            has_community_boat_facilities (bool, optional): Defaults to None
            status (Set["for_sale" | "ready_to_build" | "sold"]): Filter based on status. Defaults to ('for_sale')
            sold_date_min (datetime, optional): The sold date is min/start of range. Note that this only works if status is 'sold'. Defaults to None
            sold_date_max (datetime, optional): The sold date max/end of range. Note that this only works if status is 'sold'. Defaults to None
            prop_type (Set["farm" | "mobile" | "multi_family" | "townhomes" | "duplex_triplex" | "single_family" | "condos" | "condo_townhome_rowhome_coop" | "condo_townhome" | "land"]): The property types to include. Defaults to None (includes all)
            keywords (List[str]): List of keywords to require properties to have. Defaults to None
            tags (List[str]): List of tags to require properties to have. Defaults to None
            exlcude_tag (List[str]): List of tags require properties to NOT have. Defaults to None

        Returns:
            Sale_Query: Returns self
        """

        query = {}

        if senior_community is True:
            tags.append('senior_community')
        if single_story is True:
            tags.append('single_story')
        if multi_story is True:
            tags.append('two_or_more_stories')
        if min_garages >= 1:
            tags.append(f'garage_{min_garages}_or_more')
        if has_carport is True:
            tags.append('carport')
        if has_rv_or_boat_parking is True:
            tags.append('rv_or_boat_parking')
        if has_central_air is True:
            tags.append('central_air')
        if has_forced_air is True:
            tags.append('forced_air')
        if has_central_heat is True:
            tags.append('central_heat')
        if has_basement is True:
            tags.append('basement')
        if has_hardwood_floors is True:
            tags.append('hardwood_floors')
        if has_fireplace is True:
            tags.append('fireplace')
        if has_disability_features is True:
            tags.append('disability_features')
        if has_dining_room is True:
            tags.append('dining_room')
        if has_family_room is True:
            tags.append('family_room')
        if has_den_or_office is True:
            tags.append('den_or_office')
        if has_swimming_pool is True:
            tags.append('swimming_pool')
        if has_spa_or_hot_tub is True:
            tags.append('spa_or_hot_tub')
        if has_horse_facilities is True:
            tags.append('horse_facilities')
        if has_hill_or_mountain_view is True:
            tags.append('hill_or_mountain_view')
        if has_ocean_view is True:
            tags.append('ocean_view')
        if has_lake_view is True:
            tags.append('lake_view')
        if has_river_view is True:
            tags.append('river_view')
        if has_golf_course_lot_or_frontage is True:
            tags.append('golf_course_lot_or_frontage')
        if has_corner_lot is True:
            tags.append('corner_lot')
        if is_cul_de_sac is True:
            tags.append('cul_de_sac')
        if has_waterfront is True:
            tags.append('waterfront')
        if has_community_swimming_pool is True:
            tags.append('community_swimming_pool')
        if has_community_spa_or_hot_tub is True:
            tags.append('community_spa_or_hot_tub')
        if has_community_golf is True:
            tags.append('community_golf')
        if has_community_security_features is True:
            tags.append('community_security_features')
        if has_community_clubhouse is True:
            tags.append('community_clubhouse')
        if has_tennis_court is True:
            tags.append('tennis_court')
        if has_community_boat_facilities is True:
            tags.append('community_boat_facilities')

        if senior_community is False:
            exclude_tags.append('senior_community')

        if tags:
            query['tags'] = list(set(tags))
        if exclude_tags:
            query[exclude_tags] - list(set(exclude_tags))

        if search_location:
            query['search_location'] = {'location': search_location}

            if location_radius:
                query['search_location']['buffer'] = location_radius

        if cities:
            for city in cities:
                if 'city' not in city:
                    raise KeyError(
                        "city is a required key when specifying cities")
                if 'state_code' not in city:
                    raise KeyError(
                        'state_code is a required key when specifying cities')

            query['locations'] = cities

        if state_code:
            query['state_code'] = state_code

        query['primary'] = primary

        if new_construction is not None:
            query['new_construction'] = new_construction

        if foreclosure is not None:
            query['foreclosure'] = foreclosure

        if contingent is not None:
            query['contingent'] = contingent

        if has_tour is not None:
            query['has_tour'] = has_tour

        if open_house_date_min or open_house_date_max:
            query['open_house'] = {}
            if open_house_date_min:
                query['open_house']['min'] = open_house_date_min.strftime(
                    "%Y-%m-%d")
            if open_house_date_max:
                query['open_house']['max'] = open_house_date_max.strftime(
                    "%Y-%m-%d")

        if list_date_min or list_date_max:
            query['list_date'] = {}
            if list_date_min:
                query['list_date'] = list_date_min.strftime(
                    "%Y-%m-%d")
            if list_date_max:
                query['list_date'] = list_date_max.strftime(
                    "%Y-%m-%d")

        if price_min or price_max:
            query['listing_price'] = {}
            if price_min:
                query['listing_price']['min'] = price_min
            if price_max:
                query['listing_price']['max'] = price_max

        if price_reduce_date_min or price_reduce_date_max:
            query['price_reduced_date'] = {}
            if price_reduce_date_min:
                query['price_reduced_date']['min'] = price_reduce_date_min.strftime(
                    "%Y-%m-%d")
            if price_reduce_date_max:
                query['price_reduced_date']['max'] = price_reduce_date_max.strftime(
                    "%Y-%m-%d")

        if hoa_fee_min or hoa_fee_max:
            if include_incomplete_hoa:
                hoa_key = "hoa_fee_optional"
            else:
                hoa_key = "hoa_fee"

            query[hoa_key] = {}
            if hoa_fee_min:
                query[hoa_key]['min'] = hoa_fee_min
            if hoa_fee_max:
                query[hoa_key]['max'] = hoa_fee_max

        if interior_sqft_max or interior_sqft_min:
            query['sqft'] = {}

            if interior_sqft_min:
                query['sqft']['min'] = interior_sqft_min
            if interior_sqft_max:
                query['sqft']['max'] = interior_sqft_max

        if lot_sqft_max or lot_sqft_min:
            query['lot_sqft'] = {}

            if lot_sqft_min:
                query['lot_sqft']['min'] = lot_sqft_min
            if lot_sqft_max:
                query['lot_sqft']['max'] = lot_sqft_max

        if year_built_min or year_built_max:
            query['year_built'] = {}

            if year_built_min:
                query['year_built']['min'] = year_built_min
            if year_built_max:
                query['year_built']['max'] = year_built_max

        if bedrooms_min or bedrooms_max:
            query['beds'] = {}
            if price_min:
                query['beds']['min'] = bedrooms_min
            if price_max:
                query['beds']['max'] = bedrooms_max

        if bathrooms_min or bathrooms_max:
            query['baths'] = {}
            if price_min:
                query['baths']['min'] = bathrooms_min
            if price_max:
                query['baths']['max'] = bathrooms_max

        query['status'] = list(status)

        if sold_date_min or sold_date_max:
            query['sold_date'] = {}
            if price_min:
                query['sold_date']['min'] = sold_date_min
            if price_max:
                query['sold_date']['max'] = sold_date_max

        if prop_type:
            query['type'] = list(prop_type)

        if keywords:
            query['keywords'] = keywords

        self.payload['variables']['query'] = query
        return self

    def set_operation_name(self, name: str = "ConsumerSearchMainQuery") -> 'Sale_Query':
        """Sets the operationName in the request payload. Not recommended to change.

        Args:
            name (str, optional): operationName. Defaults to "ConsumerSearchMainQuery".

        Returns:
            Query: Returns self
        """

        if name is None:
            try:
                self.payload.pop('operationName')
            except KeyError:
                pass
        else:
            self.payload['operationName'] = name

        return self

    def set_call_from(self, call_from: str = "SRP") -> 'Sale_Query':
        """Sets the callfrom in the request payload. Not recommended to change.

        Args:
            call_from (str, optional): call_from. Defaults to "SRP".

        Returns:
            Query: Returns self
        """

        if call_from is None:
            try:
                self.payload.pop('callfrom')
            except KeyError:
                pass
        else:
            self.payload['callfrom'] = call_from

        return self

    def set_nr_query_type(self, nr_query_type: str = None) -> 'Sale_Query':
        """Sets the nrQueryType in the request payload. Not recommended to change.

        Args:
            nr_query_type (str, optional): nr_query_type. Defaults to None.

        Returns:
            Query: Returns self
        """

        if nr_query_type is None:
            try:
                self.payload.pop('nrQueryType')
            except KeyError:
                pass
        else:
            self.payload['nrQueryType'] = nr_query_type

        return self

    def set_visitor_id(self, visitor_id: str = None) -> 'Sale_Query':
        """Sets the visitor_id in the request payload. Unless set via this method, no visitor_id will be included in the payload

        Args:
            visitor_id (str, optional): visitor_id. Defaults to None.

        Returns:
            Query: Returns self
        """

        if visitor_id is None:
            try:
                self.payload.pop('visitor_id')
            except KeyError:
                pass
        else:
            self.payload['visitor_id'] = visitor_id

        return self

    def set_is_client(self, is_client: bool = True) -> 'Sale_Query':
        """Sets the isClient in the request payload. Unclear as it what this does.

        Args:
            is_client (bool, optional): isClient. Defaults to True.

        Returns:
            Query: Returns self
        """

        if is_client is None:
            try:
                self.payload.pop('isClient')
            except KeyError:
                pass
        else:
            self.payload['isClient'] = is_client

        return self

    def set_seo_payload(self, seo_payload: Dict[str, Any] = None) -> 'Sale_Query':
        """Sets the seoPayload in the request payload. It is unclear what this does. This is also by default excluded from the payload unless specifically added by this method.

        Args:
            seo_payload (Dict[str, Any], optional): seoPayload. Defaults to None.

        Returns:
            Query: Returns self
        """

        if seo_payload is None:
            try:
                self.payload.pop('seoPayload')
            except KeyError:
                pass
        else:
            self.payload['seoPayload'] = seo_payload

        return self

    def get_payload(self) -> Dict[str, Any]:
        """Returns the query POST request JSON payload.

        Returns:
            Dict[str, Any]: The query POST request JSON payload
        """
        return self.payload

    def get_post_params(self) -> Dict[str, Any]:
        """Returns the query POST request parameters.

        Returns:
            Dict[str, Any]: The query POST request paramters
        """
        return self.post_params

    def get_request(self, url: str = defaults.SEARCH_URL, headers: Dict = defaults.HEADER) -> requests.Response:
        """Sends a POST request to the Realtor.com GRAPHQL PUBLIC API and returns the response.

        Args:
            url (str, optional): The GRAPHQL API URL. Defaults to defaults.URL.
            headers (Dict, optional): The header parameters as a dictionary. Defaults to defaults.HEADER.

        Returns:
            requests.Response: The request response
        """
        return requests.post(url, json=self.payload, headers=headers, params=self.post_params)

    def get_response(self, returns: Literal["full", "results"] = "results", url: str = defaults.SEARCH_URL, headers: Dict = defaults.HEADER) -> Dict[str, Any]:
        """Sends a POST request to Realtor.com GRAPHQL Public API and returns the JSON content of the response.

        Args:
            returns (Literal[&quot;full&quot;, &quot;results&quot;], optional): This determines if the full json dictionary will be returned ("full") or just the search results ("results"). Note that the search results include the count, total, and 'results' which contains the list of acutal results. Defaults to "results".
            url (str, optional): The GRAPHQL API URL. Defaults to defaults.URL.
            headers (Dict, optional): The header parameters as a dictionary. Defaults to defaults.HEADER.

        Returns:
            Dict[str, Any]: The request response json data
        """

        data = self.get_request(url, headers).json()

        if returns == "full":
            return data
        elif returns == "results":
            return data.get('data', {}).get('home_search')
        else:
            raise ValueError(
                f"returns param must be either 'full' or 'results', was {returns}")
